import pytest
from ophyd import EpicsMotor, sim
from ophyd_async.epics.motor import Motor

from haven.device import connect_devices, make_device
from haven.devices.motor import HavenMotor


@pytest.mark.asyncio
async def test_connect_devices():
    motor = Motor("255idc:m1", name="motor")
    new_motors = await connect_devices([motor], mock=True)
    assert len(new_motors) == 1
    assert new_motors[0] is motor


@pytest.mark.asyncio
async def test_connect_devices_with_registry(sim_registry):
    motor = Motor("255idc:m1", name="motor")
    await connect_devices([motor], mock=True, labels={"motors"}, registry=sim_registry)
    # Check that we can find our devices in the registry
    sim_registry.find(name="motor")
    sim_registry.find(label="motors")


def test_load_fake_device(sim_registry):
    """Does ``make_device`` create a fake device if beamline is disconnected?"""
    motor = make_device(HavenMotor, name="real_motor")
    assert isinstance(motor.user_readback, sim.SynSignal)


def test_accept_fake_device(sim_registry):
    """Does ``make_device`` use a specific fake device if beamline is disconnected?"""
    FakeMotor = sim.make_fake_device(EpicsMotor)
    motor = make_device(HavenMotor, name="real_motor", FakeDeviceClass=FakeMotor)
    assert isinstance(motor, FakeMotor)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from haven.devices.xspress import Xspress3Detector


def test_num_elements(xspress):
    assert xspress.num_elements == 4


def test_num_rois(xspress):
    assert xspress.num_rois == 16


@pytest.mark.skip(
    reason="This test can't instantiate the device without having an IOC present"
)
def test_mca_signals():
    xsp = Xspress3Detector("255id_xsp:", name="spcxsp")
    assert not xsp.connected
    # Spot-check some PVs
    assert xsp.cam.acquire_time._write_pv.pvname == "255id_xsp:det1:AcquireTime"
    assert xsp.cam.acquire._write_pv.pvname == "255id_xsp:det1:Acquire"
    assert xsp.cam.acquire._read_pv.pvname == "255id_xsp:det1:Acquire_RBV"
    assert (
        xsp.mcas.mca0.rois.roi0.total_count._read_pv.pvname
        == "255id_xsp:MCA1ROI:1:Total_RBV"
    )


def test_roi_size(xspress):
    """Do the signals for max/size auto-update."""
    roi = xspress.mcas.mca0.rois.roi0
    roi.lo_chan.set(10).wait()
    # Update the size and check the maximum
    roi.size.set(7).wait()
    assert roi.hi_chan.get() == 17
    # Update the maximum and check the size
    roi.hi_chan.set(28).wait()
    assert roi.size.get() == 18
    # Update the minimum and check the size
    roi.lo_chan.set(25).wait()
    assert roi.size.get() == 3


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import time
from collections import OrderedDict
from enum import IntEnum
from functools import partial
from typing import Dict, Optional, Sequence

import numpy as np
import pandas as pd
from apstools.devices import CamMixin_V34, SingleTrigger_V34
from ophyd import ADComponent as ADCpt
from ophyd import Component as Cpt
from ophyd import Device
from ophyd import DynamicDeviceComponent as DDC
from ophyd import EpicsSignal, EpicsSignalRO, Kind
from ophyd.areadetector.base import EpicsSignalWithRBV as SignalWithRBV
from ophyd.signal import InternalSignal
from ophyd.status import StatusBase, SubscriptionStatus
from pcdsdevices.signal import MultiDerivedSignal
from pcdsdevices.type_hints import OphydDataType, SignalToValue

from ..device import RegexComponent as RECpt
from ..device import make_device
from .area_detector import DetectorBase, HDF5FilePlugin
from .fluorescence_detector import (
    MCASumMixin,
    ROIMixin,
    UseROISignal,
    XRFMixin,
    add_roi_sums,
)

__all__ = ["Xspress3Detector", "ROI"]


log = logging.getLogger(__name__)


active_kind = Kind.normal | Kind.config


NUM_ROIS = 16


class ChannelSignal(MultiDerivedSignal):
    """A high/low range limit channel for an ROI."""

    def set(
        self,
        value: OphydDataType,
        *,
        timeout: Optional[float] = None,
        settle_time: Optional[float] = None,
    ) -> StatusBase:
        # Check for existing signals and, if necessary, wait them out
        signals = [
            self.parent.hi_chan,
            self.parent.lo_chan,
            self.parent.size,
            self.parent._lo_chan,
        ]

        def get_threads():
            thds = [sig._set_thread for sig in signals if sig._set_thread]
            return [th for th in thds if th is not None]

        while len(threads := get_threads()) > 0:
            for th in threads:
                th.join()
        # Set the signal like normal
        return super().set(value, timeout=timeout, settle_time=settle_time)


class ROI(ROIMixin):
    def _get_hi_chan(self, mds: MultiDerivedSignal, items: SignalToValue) -> int:
        # Make sure other signals don't have pending threads
        lo = items[self._lo_chan]
        size = items[self.size]
        return lo + size

    def _put_hi_chan(
        self, mds: MultiDerivedSignal, value: OphydDataType
    ) -> SignalToValue:
        lo = self._lo_chan.get()
        new_size = value - lo
        return {self.size: new_size}

    def _get_lo_chan(self, mds: MultiDerivedSignal, items: SignalToValue) -> int:
        return items[self._lo_chan]

    def _put_lo_chan(
        self, mds: MultiDerivedSignal, value: OphydDataType
    ) -> SignalToValue:
        hi = self.hi_chan.get()
        return {
            self._lo_chan: value,
            self.size: hi - value,
        }

    label = Cpt(EpicsSignal, "Name", kind="config")
    _lo_chan = Cpt(EpicsSignal, "MinX", kind="omitted")
    size = Cpt(EpicsSignal, "SizeX", kind="config")
    hi_chan = Cpt(
        ChannelSignal,
        attrs=["_lo_chan", "size"],
        calculate_on_get=_get_hi_chan,
        calculate_on_put=_put_hi_chan,
    )
    lo_chan = Cpt(
        ChannelSignal,
        attrs=["_lo_chan", "size"],
        calculate_on_get=_get_lo_chan,
        calculate_on_put=_put_lo_chan,
    )
    background_width = Cpt(EpicsSignal, "BgdWidth", kind="config")
    use = Cpt(UseROISignal, derived_from="label", kind="config")

    count = Cpt(EpicsSignalRO, "Total_RBV", kind="normal")
    net_count = Cpt(EpicsSignalRO, "Net_RBV", kind="normal")
    min_count = Cpt(EpicsSignalRO, "MinValue_RBV", kind="normal")
    max_count = Cpt(EpicsSignalRO, "MaxValue_RBV", kind="normal")
    mean_count = Cpt(EpicsSignalRO, "MeanValue_RBV", kind="normal")

    _default_read_attrs = [
        "count",
        "net_count",
    ]
    _default_configuration_attrs = [
        "label",
        "background_width",
        "hi_chan",
        "lo_chan",
    ]
    kind = active_kind


def add_rois(
    range_: Sequence[int] = range(NUM_ROIS), kind=Kind.normal, lazy=True, **kwargs
):
    """Add one or more ROIs to an MCA instance

    Parameters
    ----------
    range_ : sequence of ints
        Must be be in the set [0,31]

    By default, an EpicsMCA is initialized with all 32 rois. These
    provide the following Components as EpicsSignals (N=[0,31]):
    EpicsMCA.rois.roiN.(label,count,net_count,preset_cnt, is_preset,
    bkgnd_chans, hi_chan, lo_chan)

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    kwargs["lazy"] = lazy
    for roi in range_:
        if not (0 <= roi <= 47):
            raise ValueError(f"roi {roi} must be in the set [0,47]")
        attr = f"roi{roi}"
        defn[attr] = (
            ROI,
            f"ROI:{roi+1}:",
            kwargs,
        )
    return defn


class MCARecord(MCASumMixin, Device):
    rois = DDC(add_rois(), kind=active_kind)
    spectrum = ADCpt(EpicsSignalRO, ":ArrayData", kind="normal", lazy=True)
    dead_time_percent = RECpt(
        EpicsSignalRO,
        ":DeadTime_RBV",
        pattern=r":MCA",
        repl=":C",
        lazy=True,
        kind="normal",
    )
    dead_time_factor = RECpt(
        EpicsSignalRO,
        ":DTFactor_RBV",
        pattern=r":MCA",
        repl=":C",
        lazy=True,
        kind="normal",
    )
    clock_ticks = RECpt(
        EpicsSignalRO,
        "SCA:0:Value_RBV",
        pattern=r":MCA",
        repl=":C",
        lazy=True,
        kind="normal",
    )
    _default_read_attrs = [
        "rois",
        "spectrum",
        "dead_time_percent",
        "dead_time_factor",
        "clock_ticks",
        "total_count",
    ]
    _default_configuration_attrs = ["rois"]
    kind = active_kind


def add_mcas(range_, kind=active_kind, **kwargs):
    """Add one or more MCARecords to a device

    Parameters
    ----------
    range_
      Indices for which to create MCA records.

    """
    defn = OrderedDict()
    kwargs["kind"] = kind
    kwargs["lazy"] = True
    for idx in range_:
        attr = f"mca{idx}"
        defn[attr] = (
            MCARecord,
            f"MCA{idx+1}",
            kwargs,
        )
    return defn


class Xspress3Detector(SingleTrigger_V34, DetectorBase, XRFMixin):
    """A fluorescence detector plugged into an Xspress3 readout."""

    _dead_times: dict

    def get_acquire_frames(self, mds: MultiDerivedSignal, items: SignalToValue) -> int:
        return items[self.acquire]

    def put_acquire_frames(
        self, mds: MultiDerivedSignal, value: OphydDataType, num_frames: int
    ) -> SignalToValue:
        return {
            self.cam.num_images: num_frames,
            self.acquire: value,
        }

    cam = ADCpt(CamMixin_V34, "det1:")
    hdf = ADCpt(HDF5FilePlugin, "HDF1:", kind=Kind.normal)
    # Core control interface signals
    detector_state = ADCpt(EpicsSignalRO, "det1:DetectorState_RBV", kind="omitted")
    acquire = ADCpt(SignalWithRBV, "det1:Acquire", kind="omitted")
    acquire_busy = ADCpt(EpicsSignalRO, "det1:AcquireBusy", kind="omitted")
    acquire_period = ADCpt(SignalWithRBV, "det1:AcquirePeriod", kind="omitted")
    dwell_time = ADCpt(SignalWithRBV, "det1:AcquireTime", kind="normal")
    erase = ADCpt(EpicsSignal, "det1:ERASE", kind="omitted")
    acquire_single = ADCpt(
        MultiDerivedSignal,
        attrs=["acquire", "cam.num_images"],
        calculate_on_get=get_acquire_frames,
        calculate_on_put=partial(put_acquire_frames, num_frames=1),
    )
    acquire_multiple = ADCpt(
        MultiDerivedSignal,
        attrs=["acquire", "cam.num_images"],
        calculate_on_get=get_acquire_frames,
        calculate_on_put=partial(put_acquire_frames, num_frames=2000),
    )
    # Dead time aggregate statistics
    dead_time_average = ADCpt(InternalSignal, kind="normal")
    dead_time_min = ADCpt(InternalSignal, kind="normal")
    dead_time_max = ADCpt(InternalSignal, kind="normal")

    # Number of elements is overridden by subclasses
    mcas = DDC(
        add_mcas(range_=range(1)),
        kind=active_kind,
        default_read_attrs=["mca0"],
        default_configuration_attrs=["mca0"],
    )
    roi_sums = DDC(
        add_roi_sums(mcas=range(1), rois=range(NUM_ROIS)),
        kind=active_kind,
        default_read_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
        default_configuration_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
    )

    _default_read_attrs = [
        "cam",
        "mcas",
        "roi_sums",
        "dead_time_average",
        "dead_time_min",
        "dead_time_max",
    ]
    _default_configuration_attrs = [
        "cam",
        "mcas",
    ]

    class erase_states(IntEnum):
        DONE = 0
        ERASE = 1

    class acquire_states(IntEnum):
        DONE = 0
        ACQUIRE = 1

    class detector_states(IntEnum):
        IDLE = 0
        ACQUIRE = 1
        READOUT = 2
        CORRECT = 3
        SAVING = 4
        ABORTING = 5
        ERROR = 6
        WAITING = 7
        INITIALIZING = 8
        DISCONNECTED = 9
        ABORTED = 10

    class trigger_modes(IntEnum):
        SOFTWARE = 0
        INTERNAL = 1
        IDC = 2
        TTL_VETO_ONLY = 3
        TTL_BOTH = 4
        LVDS_VETO_ONLY = 5
        LVDS_BOTH = 6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stage_sigs[self.cam.trigger_mode] = self.trigger_modes.INTERNAL
        self.stage_sigs[self.cam.num_images] = 1
        # The image mode is not a real signal in the Xspress3 IOC
        del self.stage_sigs["cam.image_mode"]
        # Set up subscriptions for dead-time calculations
        self._dead_times = {}
        for mca in self.mca_records():
            mca.dead_time_percent.subscribe(self._update_dead_time_calcs)

    @property
    def default_time_signal(self):
        return self.cam.acquire_time

    def _update_dead_time_calcs(self, *args, value, obj, **kwargs):
        self._dead_times[obj.name] = value
        # Calculate aggregate dead time stats
        dead_times = np.asarray(list(self._dead_times.values()))
        self.dead_time_average.put(float(np.mean(dead_times)), internal=True)
        self.dead_time_min.put(float(np.min(dead_times)), internal=True)
        self.dead_time_max.put(float(np.max(dead_times)), internal=True)

    @property
    def stage_num_frames(self):
        """How many frames to prepare for when staging this detector."""
        return self.stage_sigs.get(self.num_frames, 1)

    @stage_num_frames.setter
    def stage_num_frames(self, val):
        self.stage_sigs[self.num_frames] = val

    @property
    def num_elements(self):
        return len(self.mcas.component_names)

    def save_fly_datum(self, *, value, timestamp, obj, **kwargs):
        """Callback to save data from a signal during fly-scanning."""
        datum = (timestamp, value)
        self._fly_data.setdefault(obj, []).append(datum)

    def fly_data(self):
        """Compile the fly-scan data into a pandas dataframe.

        Some stray rows show up at the beginning that get dropped:

        - Old image counter with previous values capturing when
          subscribing. Needed to make sure we have an entry for all
          the signals.
        - The counts collected during taxiing.

        """
        # Get the data for frame number as a reference
        image_counter = pd.DataFrame(
            self._fly_data[self.cam.array_counter],
            columns=["timestamps", "image_counter"],
        )
        image_counter["image_counter"] -= 2  # Correct for stray frames
        # Build all the individual signals' dataframes
        dfs = []
        for sig, data in self._fly_data.items():
            df = pd.DataFrame(data, columns=["timestamps", sig])
            old_shape = df.shape
            nums = (df.timestamps - image_counter.timestamps).abs()

            # Assign each datum an image number based on timestamp
            def get_image_num(ts):
                """Get the image number taken closest to a given timestamp."""
                num = image_counter.iloc[
                    (image_counter["timestamps"] - ts).abs().argsort()[:1]
                ]
                num = num["image_counter"].iloc[0]
                return num

            im_nums = [get_image_num(ts) for ts in df.timestamps.values]
            df.index = im_nums
            # Remove duplicates and intermediate ROI sums
            df.sort_values("timestamps")
            df = df.groupby(df.index).last()
            dfs.append(df)
        # Combine frames into monolithic dataframes
        data = image_counter.copy()
        data = data.set_index("image_counter", drop=True)
        timestamps = data.copy()
        for df in dfs:
            sig = df.columns[1]
            data[sig] = df[sig]
            timestamps[sig] = df["timestamps"]
        # Fill in missing values, most likely because the value didn't
        # change so no new camonitor reply was received
        data = data.ffill(axis=0)
        timestamps = timestamps.ffill(axis=1)
        # Drop the extra rows that come from the subscription setup
        data = data.iloc[1:]
        timestamps = timestamps.iloc[1:]
        # Drop extra rows from before and during taxi
        for idx in [-2, -1]:
            try:
                data.drop(idx, inplace=True)
                timestamps.drop(idx, inplace=True)
            except KeyError:
                continue
        return data, timestamps

    def walk_fly_signals(self, *, include_lazy=False):
        """Walk all signals in the Device hierarchy that are to be read during
        fly-scanning.

        Parameters
        ----------
        include_lazy : bool, optional
            Include not-yet-instantiated lazy signals

        Yields
        ------
        ComponentWalk
            Where ancestors is all ancestors of the signal, including the
            top-level device `walk_signals` was called on.

        """
        for walk in self.walk_signals():
            # Image counter has to be included for data alignment
            if walk.item is self.cam.array_counter:
                yield walk
                continue
            # Only include readable signals
            if not bool(walk.item.kind & Kind.normal):
                continue
            # ROI sums do not get captured properly during flying
            # Instead, they should be calculated at the end
            # if self.roi_sums in walk.ancestors:
            #     continue
            yield walk

    def kickoff(self) -> StatusBase:
        # Set up subscriptions for capturing data
        self._fly_data = {}
        for walk in self.walk_fly_signals():
            sig = walk.item
            sig.subscribe(self.save_fly_datum, run=True)

        # Set up the status for when the detector is ready to fly
        def check_acquiring(*, old_value, value, **kwargs):
            is_acquiring = value == self.detector_states.ACQUIRE
            if is_acquiring:
                self.start_fly_timestamp = time.time()
            return is_acquiring

        status = SubscriptionStatus(self.detector_state, check_acquiring)
        # Set the right parameters
        status &= self.cam.trigger_mode.set(self.trigger_modes.TTL_VETO_ONLY)
        status &= self.cam.num_images.set(2**14)
        status &= self.acquire.set(self.acquire_states.ACQUIRE)
        return status

    def complete(self) -> StatusBase:
        """Wait for flying to be complete.

        This commands the Xspress to stop acquiring fly-scan data.

        Returns
        -------
        complete_status : StatusBase
          Indicate when flying has completed
        """
        # Remove subscriptions for capturing fly-scan data
        for walk in self.walk_fly_signals():
            sig = walk.item
            sig.clear_sub(self.save_fly_datum)
        return self.acquire.set(0)

    def collect(self) -> dict:
        """Generate the data events that were collected during the fly scan."""
        # Load the collected data, and get rid of extras
        fly_data, fly_ts = self.fly_data()
        fly_data.drop("timestamps", inplace=True, axis="columns")
        fly_ts.drop("timestamps", inplace=True, axis="columns")
        # Yield each row one at a time
        for data_row, ts_row in zip(fly_data.iterrows(), fly_ts.iterrows()):
            payload = {
                "data": {sig.name: val for (sig, val) in data_row[1].items()},
                "timestamps": {sig.name: val for (sig, val) in ts_row[1].items()},
                "time": float(np.median(np.unique(ts_row[1].values))),
            }
            yield payload

    def describe_collect(self) -> Dict[str, Dict]:
        """Describe details for the flyer collect() method"""
        desc = OrderedDict()
        for walk in self.walk_fly_signals():
            desc.update(walk.item.describe())
        return {self.name: desc}


def make_xspress_device(name, prefix, num_elements):
    # Build the mca components
    # (Epics uses 1-index instead of 0-index)
    mca_range = range(num_elements)
    attrs = {
        "mcas": DDC(
            add_mcas(range_=mca_range),
            kind=active_kind,
            default_read_attrs=[f"mca{i}" for i in mca_range],
            default_configuration_attrs=[f"mca{i}" for i in mca_range],
        ),
        "roi_sums": DDC(
            add_roi_sums(mcas=mca_range, rois=range(NUM_ROIS)),
            kind=active_kind,
            default_read_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
            default_configuration_attrs=[f"roi{i}" for i in range(NUM_ROIS)],
        ),
    }
    # Create a dynamic subclass with the MCAs
    class_name = name.title().replace("_", "")
    parent_classes = (Xspress3Detector,)
    Cls = type(class_name, parent_classes, attrs)
    return make_device(
        Cls,
        name=name,
        prefix=f"{prefix}:",
        labels={"xrf_detectors", "fluorescence_detectors", "detectors"},
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""This is a copy of the apstools Slits support with signals for the tweak PV."""

import logging

from apstools.devices import PVPositionerSoftDone
from apstools.synApps.db_2slit import Optics2Slit1D, Optics2Slit2D_HV
from ophyd import Component as Cpt
from ophyd import DerivedSignal, Device, EpicsSignal
from ophyd import FormattedComponent as FCpt

from .motor import HavenMotor

log = logging.getLogger(__name__)


class PVPositionerWithTweaks(PVPositionerSoftDone):
    user_readback = Cpt(DerivedSignal, derived_from="readback")
    user_setpoint = Cpt(DerivedSignal, derived_from="setpoint")
    tweak_value = FCpt(EpicsSignal, "{prefix}{_setpoint_pv}_tweakVal.VAL")
    tweak_forward = FCpt(EpicsSignal, "{prefix}{_setpoint_pv}_tweak.B")
    tweak_reverse = FCpt(EpicsSignal, "{prefix}{_setpoint_pv}_tweak.A")


class BladePair(Optics2Slit1D):
    """
    EPICS synApps optics 2slit.db 1D support: xn, xp, size, center, sync

    "sync" is used to tell the EPICS 2slit database to synchronize the
    virtual slit values with the actual motor positions.
    """

    # Override these components to include the tweak signals
    xn = Cpt(PVPositionerWithTweaks, "", setpoint_pv="xn", readback_pv="t2.B")
    xp = Cpt(PVPositionerWithTweaks, "", setpoint_pv="xp", readback_pv="t2.A")
    size = Cpt(PVPositionerWithTweaks, "", setpoint_pv="size", readback_pv="t2.C")
    center = Cpt(PVPositionerWithTweaks, "", setpoint_pv="center", readback_pv="t2.D")


class BladeSlits(Optics2Slit2D_HV):
    """Set of slits with blades that move in and out to control beam size."""

    h = Cpt(BladePair, "H")
    v = Cpt(BladePair, "V")

    def __init__(self, prefix: str, name: str, labels={"slits"}, **kwargs):
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)


class SlitMotor(HavenMotor):
    """An Ophyd device for a motor on a set of slits.

    Similar to a regular motor with extra signals to give it the same
    interface as a non-motor based slit parameter. Different
    implementations of the slits support either provide pseudo motors,
    or a different kind of record.

    """

    readback = Cpt(DerivedSignal, derived_from="user_readback")
    setpoint = Cpt(DerivedSignal, derived_from="user_setpoint")


class ApertureSlits(Device):
    """A rotating aperture that functions like a set of slits.

    Unlike the blades slits, there are no independent parts to move,
    so each axis only has center and size.

    Based on the 25-ID-A whitebeam slits.

    The motor parameters listed below specify which motor records
    control which axis. The last piece of the PV prefix will be
    removed, and the motor number added on. For example, if the prefix
    is "255ida:slits:US:", and the pitch motor is "255ida:slits:m3",
    then *pitch_motor* should be "m3".

    Parameters
    ==========
    pitch_motor
      The motor record suffix controlling the real pitch motor. Don't
      include a field. E.g. "m3"
    yaw_motor
      The motor record suffix controlling the real yaw motor. Don't
      include a field. E.g. "m3"
    horizontal_motor
      The motor record suffix controlling the real horizontal
      motor. This is different from the horizontal slits
      position. Don't include a field. E.g. "m3"
    diagonal_motor
      The motor record suffix controlling the real diagonal
      motor. Don't include a field. E.g. "m3"

    """

    def __init__(
        self,
        prefix: str,
        name: str,
        pitch_motor: str,
        yaw_motor: str,
        horizontal_motor: str,
        diagonal_motor: str,
        labels={"slits"},
        **kwargs,
    ):
        # Determine the prefix for the motors
        pieces = prefix.strip(":").split(":")
        self.motor_prefix = ":".join(pieces[:-1])
        self._pitch_motor = pitch_motor
        self._yaw_motor = yaw_motor
        self._horizontal_motor = horizontal_motor
        self._diagonal_motor = diagonal_motor
        super().__init__(prefix=prefix, name=name, labels=labels, **kwargs)

    class SlitAxis(Device):
        size = Cpt(SlitMotor, "Size", labels={"motors"})
        center = Cpt(SlitMotor, "Center", labels={"motors"})

    # Individual slit directions
    h = Cpt(SlitAxis, "h")
    v = Cpt(SlitAxis, "v")

    # Real motors that directly control the slits
    pitch = FCpt(
        SlitMotor, "{self.motor_prefix}:{self._pitch_motor}", labels={"motors"}
    )
    yaw = FCpt(SlitMotor, "{self.motor_prefix}:{self._yaw_motor}", labels={"motors"})
    horizontal = FCpt(
        SlitMotor, "{self.motor_prefix}:{self._horizontal_motor}", labels={"motors"}
    )
    diagonal = FCpt(
        SlitMotor, "{self.motor_prefix}:{self._diagonal_motor}", labels={"motors"}
    )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
from unittest import mock

import pytest
from bluesky_queueserver_api import BPlan
from ophyd_async.core import set_mock_value
from qtpy import QtCore

from firefly.plans.grid_scan import GridScanDisplay


@pytest.fixture()
async def display(qtbot, sim_registry, sync_motors, async_motors, dxp, ion_chamber):
    display = GridScanDisplay()
    qtbot.addWidget(display)
    await display.update_devices(sim_registry)
    display.ui.run_button.setEnabled(True)
    try:
        yield display
    finally:
        await asyncio.sleep(0.1)


@pytest.mark.asyncio
async def test_time_calculator(display, sim_registry, ion_chamber):
    # set up motor num
    await display.update_regions(2)

    # set up num of repeat scans
    display.ui.spinBox_repeat_scan_num.setValue(6)

    # set up scan num of points
    display.regions[0].scan_pts_spin_box.setValue(4)
    display.regions[1].scan_pts_spin_box.setValue(5)

    # set up detectors
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", ion_chamber.name]
    )

    # set up default timing for the detector
    detectors = display.ui.detectors_list.selected_detectors()
    detectors = {name: sim_registry[name] for name in detectors}
    set_mock_value(detectors[ion_chamber.name].default_time_signal, 0.82)
    detectors["vortex_me4"].default_time_signal.set(0.5).wait()

    # Run the time calculator
    await display.update_total_time()

    # Check whether time is calculated correctly for a single scan
    assert display.ui.label_hour_scan.text() == "0"
    assert display.ui.label_min_scan.text() == "0"
    assert display.ui.label_sec_scan.text() == "16.4"

    # Check whether time is calculated correctly including the repeated scan
    assert display.ui.label_hour_total.text() == "0"
    assert display.ui.label_min_total.text() == "1"
    assert display.ui.label_sec_total.text() == "38.4"


@pytest.mark.asyncio
async def test_grid_scan_plan_queued(display, qtbot, sim_registry, ion_chamber):
    await display.update_regions(2)

    # set up a test motor 1
    display.regions[0].motor_box.combo_box.setCurrentText("sync_motor_2")
    display.regions[0].start_line_edit.setText("1")
    display.regions[0].stop_line_edit.setText("111")
    display.regions[0].scan_pts_spin_box.setValue(5)

    # select snake for the first motor
    display.regions[0].snake_checkbox.setChecked(True)

    # set up a test motor 2
    display.regions[1].motor_box.combo_box.setCurrentText("async_motor_1")
    display.regions[1].start_line_edit.setText("2")
    display.regions[1].stop_line_edit.setText("222")
    display.regions[1].scan_pts_spin_box.setValue(10)

    # set up detector list
    display.ui.detectors_list.selected_detectors = mock.MagicMock(
        return_value=["vortex_me4", ion_chamber.name]
    )
    # set up meta data
    display.ui.lineEdit_sample.setText("sam")
    display.ui.lineEdit_purpose.setText("test")
    display.ui.textEdit_notes.setText("notes")

    expected_item = BPlan(
        "grid_scan",
        ["vortex_me4", "I00"],
        "async_motor_1",
        2.0,
        222.0,
        10,
        "sync_motor_2",
        1.0,
        111.0,
        5,
        snake_axes=["sync_motor_2"],
        md={"sample": "sam", "purpose": "test", "notes": "notes"},
    )

    def check_item(item):
        print(item.to_dict())
        print(expected_item.to_dict())
        return item.to_dict() == expected_item.to_dict()

    # Click the run button and see if the plan is queued
    with qtbot.waitSignal(
        display.queue_item_submitted, timeout=1000, check_params_cb=check_item
    ):
        qtbot.mouseClick(display.ui.run_button, QtCore.Qt.LeftButton)



import logging
import warnings

from apstools.utils.misc import safe_ophyd_name
from ophyd import Component as Cpt
from ophyd import EpicsMotor, EpicsSignal, EpicsSignalRO, Kind
from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    ConfigSignal,
    SignalBackend,
    SignalX,
    SubsetEnum,
)
from ophyd_async.epics.motor import Motor as MotorBase
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw
from ophyd_async.epics.signal._signal import _epics_signal_backend
from ophydregistry import Registry

from .motor_flyer import MotorFlyer

log = logging.getLogger(__name__)


class SignalX(SignalX):
    trigger_value = 1

    def __init__(self, *args, trigger_value=1, **kwargs):
        self.trigger_value = trigger_value
        super().__init__(*args, **kwargs)

    def trigger(
        self, wait=False, timeout: CalculatableTimeout = CALCULATE_TIMEOUT
    ) -> AsyncStatus:
        """Trigger the action and return a status saying when it's done"""
        if timeout is CALCULATE_TIMEOUT:
            timeout = self._timeout
        coro = self._backend.put(self.trigger_value, wait=wait, timeout=timeout)
        return AsyncStatus(coro)


def epics_signal_x(write_pv: str, name: str = "", trigger_value=1) -> SignalX:
    """Create a `SignalX` backed by 1 EPICS PVs

    Parameters
    ----------
    write_pv:
      The PV to write its initial value to on trigger
    trigger_value:
      The value to send to the write PV.
    """
    backend: SignalBackend = _epics_signal_backend(None, write_pv, write_pv)
    return SignalX(backend, name=name, trigger_value=trigger_value)


class Motor(MotorBase):
    """The default motor for asynchrnous movement."""

    def __init__(
        self, prefix: str, name="", labels={"motors"}, auto_name: bool = None
    ) -> None:
        """Parameters
        ==========
        auto_name
          If true, or None when no name was provided, the name for
          this motor will be set based on the motor's *description*
          field.

        """
        self._ophyd_labels_ = labels
        self._old_flyer_velocity = None
        self.auto_name = auto_name
        # Configuration signals
        with self.add_children_as_readables(ConfigSignal):
            self.description = epics_signal_rw(str, f"{prefix}.DESC")
            self.user_offset = epics_signal_rw(float, f"{prefix}.OFF")
            self.user_offset_dir = epics_signal_rw(
                SubsetEnum["Pos", "Neg"], f"{prefix}.DIR"
            )
            self.offset_freeze_switch = epics_signal_rw(
                SubsetEnum["Variable", "Frozen"], f"{prefix}.FOFF"
            )
        # Motor status signals
        self.motor_is_moving = epics_signal_r(int, f"{prefix}.MOVN")
        self.motor_done_move = epics_signal_r(int, f"{prefix}.DMOV")
        self.high_limit_switch = epics_signal_r(int, f"{prefix}.HLS")
        self.low_limit_switch = epics_signal_r(int, f"{prefix}.LLS")
        self.high_limit_travel = epics_signal_rw(float, f"{prefix}.HLM")
        self.low_limit_travel = epics_signal_rw(float, f"{prefix}.LLM")
        self.direction_of_travel = epics_signal_r(int, f"{prefix}.TDIR")
        self.soft_limit_violation = epics_signal_r(int, f"{prefix}.LVIO")
        # Load all the parent signals
        super().__init__(prefix=prefix, name=name)
        # Override the motor stop signal to use the right trigger value
        self.motor_stop = epics_signal_x(f"{prefix}.STOP")
        self.set_name(self.name)

    async def connect(
        self,
        mock: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
        force_reconnect: bool = False,
    ):
        """Connect self and all child Devices.

        Contains a timeout that gets propagated to child.connect methods.

        Parameters
        ----------
        mock:
            If True then use ``MockSignalBackend`` for all Signals
        timeout:
            Time to wait before failing with a TimeoutError.

        """
        await super().connect(
            mock=mock, timeout=timeout, force_reconnect=force_reconnect
        )
        # Update the device's name
        auto_name = bool(self.auto_name) or (self.auto_name is None and self.name == "")
        if bool(auto_name):
            try:
                desc = await self.description.get_value()
            except Exception as exc:
                warnings.warn(
                    f"Could not read description for {self}. " "Name not updated. {exc}"
                )
                return
            # Only update the name if the description has been set
            if desc != "":
                self.set_name(safe_ophyd_name(desc))


class HavenMotor(MotorFlyer, EpicsMotor):
    """The default motor for haven movement.

    This motor also implements the flyer interface and so can be used
    in a fly scan, though no hardware triggering is supported.

    Returns to the previous value when being unstaged.

    """

    # Extra motor record components
    encoder_resolution = Cpt(EpicsSignal, ".ERES", kind=Kind.config)
    description = Cpt(EpicsSignal, ".DESC", kind="omitted")
    tweak_value = Cpt(EpicsSignal, ".TWV", kind="omitted")
    tweak_forward = Cpt(EpicsSignal, ".TWF", kind="omitted", tolerance=2)
    tweak_reverse = Cpt(EpicsSignal, ".TWR", kind="omitted", tolerance=2)
    motor_stop = Cpt(EpicsSignal, ".STOP", kind="omitted", tolerance=2)
    soft_limit_violation = Cpt(EpicsSignalRO, ".LVIO", kind="omitted")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def stage(self):
        super().stage()
        # Override some additional staged signals
        self._original_vals.setdefault(self.user_setpoint, self.user_readback.get())
        self._original_vals.setdefault(self.velocity, self.velocity.get())


def load_motors(
    prefix: str,
    num_motors: int,
    auto_name: bool = True,
    registry: Registry | None = None,
) -> list:
    """Load generic hardware motors from IOCs.

    For example, if *prefix* is "255idcVME:" and *num_motors* is 12,
    then motors with PVs from "255idcVME:m1" to "255idcVME:m12" will
    be returned.

    Parameters
    ==========
    prefix
      The PV prefix for all motors on this IOC
    num_motors
      How many motors to create for this IOC.
    auto_name
      If true, the name of the device will be updated to match the
      motor's ``.DESC`` field.

    Returns
    =======
    devices
      The newly created motor devices.

    """
    # Create the motor devices
    devices = []
    for idx in range(num_motors):
        labels = {"motors", "extra_motors", "baseline"}
        default_name = f"{prefix.strip(':')}_m{idx+1}"
        new_motor = Motor(
            prefix=f"{prefix}m{idx+1}",
            name=default_name,
            labels=labels,
            auto_name=auto_name,
        )
        devices.append(new_motor)
    # Removed motors that are already available somewhere else (e.g. KB Mirrors)
    if registry is not None:
        existing_motors = registry.findall(label="motors", allow_none=True)
        existing_sources = [
            getattr(m.user_readback, "source", "") for m in existing_motors
        ]
        existing_sources = [s for s in existing_sources if s != ""]
        devices = [
            m
            for m in devices
            if getattr(m.user_readback, "source", "") not in existing_sources
        ]
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""A PyDM data plugin that uses an Ophyd registry object to
communicate with signals.

Provides funcitonality so that PyDM channels can be addressed as e.g.
``haven://mirror.pitch.user_readback``

"""

import asyncio
import inspect
import logging
from typing import Mapping

import numpy as np
from bluesky.protocols import Movable, Subscribable
from ophyd import OphydObject
from ophyd.utils.epics_pvs import AlarmSeverity
from pydm.data_plugins.plugin import PyDMConnection
from qasync import asyncSlot
from typhos.plugins.core import SignalConnection, SignalPlugin

from haven import beamline

from .exceptions import UnknownOphydSignal

log = logging.getLogger(__name__)


class RegistryConnection:
    def find_signal(self, address: str):
        """Find a signal in the registry given its address.
        This method is intended to be overridden by subclasses that
        may use a different mechanism to keep track of signals.
        Parameters
        ----------
        address
          The connection address for the signal. E.g. in
          "sig://sim_motor.user_readback" this would be the
          "sim_motor.user_readback" portion.
        Returns
        -------
        Signal
          The Ophyd signal corresponding to the address.
        """
        return beamline.registry[address]


class HavenConnection(RegistryConnection, SignalConnection):
    pass


class HavenAsyncConnection(RegistryConnection, PyDMConnection):
    _is_ready = False

    def __init__(self, channel, address, protocol=None, parent=None):
        # Create base connection
        super().__init__(channel, address, protocol=protocol, parent=parent)
        self._connection_open = True
        self.signal_type = None
        self.is_float = False
        # Collect our signal
        self.signal = self.find_signal(address)
        self.is_triggerable = hasattr(self.signal, "trigger")
        self.is_movable = isinstance(self.signal, Movable)
        self.is_writable = self.is_movable or self.is_triggerable
        self.is_subscribable = isinstance(self.signal, Subscribable)
        # Subscribe to updates from Ophyd
        if self.is_subscribable:
            self.signal.subscribe(self.send_new_value)
        self._is_ready = True
        # Add listener
        self.add_listener(channel)

    def add_listener(self, channel):
        super().add_listener(channel)
        # If the channel is used for writing to PVs, hook it up to the 'put' methods.
        if channel.value_signal is not None:
            for type_ in [str, int, float, np.ndarray]:
                try:
                    channel.value_signal[type_].connect(self.put_value)
                except KeyError:
                    pass
        # Emit updated value/metadata so the new channel gets notified
        self._meta_task = asyncio.create_task(
            self.send_new_meta(), name=f"meta_{self.signal.name}"
        )

    async def send_new_meta(self):
        # Assume the signal is connected
        self.connection_state_signal.emit(True)
        # Check the bluesky interface for writability
        self.write_access_signal.emit(self.is_writable)
        # Get some more metadata
        if hasattr(self.signal, "describe"):
            description = await self.signal.describe()
            description = description[self.signal.name]
        else:
            description = {}
        # What is the precision of this signal
        if (precision := description.get("precision")) is not None:
            self.prec_signal.emit(precision)
        # What are the units?
        if (units := description.get("units")) is not None:
            self.unit_signal.emit(units)
        # Update choices for enumerated types
        if (enum_strs := description.get("choices")) is not None:
            self.enum_strings_signal.emit(tuple(enum_strs))
        # Send the current value as well so widgets can update
        if self.is_subscribable:
            self.signal._get_cache()._notify(self.send_new_value, want_value=False)
        elif self.is_triggerable:
            # Any value will do, we won't use it anyway
            self.new_value_signal.emit(0)

    def send_new_value(self, reading: Mapping = {}, **kwargs):
        """
        Update the UI with a new value from the Signal.
        """
        # Ignore the run when ``subscribe()`` is first called
        if not self._is_ready:
            return
        # Update value
        reading = reading[self.signal.name]
        if reading is None:
            return
        value = reading["value"]
        try:
            self.new_value_signal[type(value)].emit(value)
        except Exception:
            log.exception("Unable to update %r with value %r.", self.signal.name, value)
        # Update alarm severity
        severity = reading.get("alarm_severity", AlarmSeverity.NO_ALARM)
        self.new_severity_signal.emit(severity)

    def close(self):
        """Unsubscribe from the Ophyd signal."""
        if self.is_subscribable:
            self.signal.clear_sub(self.send_new_value)

    @asyncSlot(int)
    @asyncSlot(float)
    @asyncSlot(str)
    @asyncSlot(np.ndarray)
    async def put_value(self, new_value):
        if self.is_triggerable:
            # Just trigger the signal and be done
            await self.signal.trigger(wait=False)
        else:
            # Put the proper value to the signal
            old_value = await self.signal.get_value()
            log.info(
                f"Moving signal '{self.signal.name}' from {old_value} to {new_value}"
            )
            await self.signal.set(new_value, wait=False)


class HavenPlugin(SignalPlugin):
    protocol = "haven"

    @staticmethod
    def connection_class(channel, address, protocol):
        # Check if we need the synchronous or asynchronous version
        try:
            sig = beamline.registry[address]
        except KeyError:
            sig = None
        is_ophyd_async = inspect.iscoroutinefunction(getattr(sig, "connect", None))
        is_vanilla_ophyd = isinstance(sig, OphydObject)
        # Get the right Connection class and build it
        if is_ophyd_async:
            return HavenAsyncConnection(channel, address, protocol)
        elif is_vanilla_ophyd:
            return HavenConnection(channel, address, protocol)
        else:
            msg = f"Signal must be ophyd or ophyd_async signal. Got {type(sig)}."
            raise UnknownOphydSignal(msg)



from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest

from haven.devices.ion_chamber import IonChamber
from haven.devices.motor import load_motors
from haven.devices.slits import BladeSlits
from haven.instrument import Instrument

haven_dir = Path(__file__).parent.parent.resolve()
toml_file = haven_dir / "iconfig_testing.toml"


@pytest.fixture()
def instrument():
    inst = Instrument(
        {
            "ion_chamber": IonChamber,
            "motors": load_motors,
            "blade_slits": BladeSlits,
        }
    )
    with open(toml_file, mode="tr", encoding="utf-8") as fd:
        inst.parse_toml_file(fd)
    return inst


def test_global_parameters(instrument):
    """Check that we loaded keys that apply to the whole beamline."""
    assert instrument.beamline_name == "SPC Beamline (sector unknown)"
    assert instrument.hardware_is_present == False


def test_validate_missing_params(instrument):
    defn = {
        # "scaler_prefix": "scaler_1:",
        # "scaler_channel": 3,
        # "preamp_prefix": "preamp_1:",
        # "voltmeter_prefix": "labjack_1:",
        # "voltmeter_channel": 1,
        # "counts_per_volt_second": 1e-6,
        # "name": "",
        # "auto_name": None,
    }
    with pytest.raises(Exception):
        instrument.validate_params(defn, IonChamber)


def test_validate_optional_params(instrument):
    defn = {
        "scaler_prefix": "scaler_1:",
        "scaler_channel": 3,
        "preamp_prefix": "preamp_1:",
        "voltmeter_prefix": "labjack_1:",
        "voltmeter_channel": 1,
        "counts_per_volt_second": 1e-6,
        # "name": "",
        # "auto_name": None,
    }
    instrument.validate_params(defn, IonChamber)


def test_validate_wrong_types(instrument):
    defn = {
        "scaler_prefix": "scaler_1:",
        "scaler_channel": "3",
        "preamp_prefix": "preamp_1:",
        "voltmeter_prefix": "labjack_1:",
        "voltmeter_channel": "1",
        "counts_per_volt_second": 1e-6,
        "name": "",
        "auto_name": None,
    }
    with pytest.raises(Exception):
        instrument.validate_params(defn, IonChamber)


async def test_connect(instrument):
    async_devices = [d for d in instrument.devices if hasattr(d, "_connect_task")]
    sync_devices = [d for d in instrument.devices if hasattr(d, "connected")]
    assert len(async_devices) > 0
    assert len(sync_devices) > 0
    # Are devices disconnected to start with?
    assert all([d._connect_task is None for d in async_devices])
    assert all([not d.connected is None for d in sync_devices])
    # Connect the device
    await instrument.connect(mock=True)
    # Are devices connected afterwards?
    # NB: This doesn't actually test the code for threaded devices
    assert all([d._connect_task.done for d in async_devices])


async def test_load(monkeypatch):
    instrument = Instrument({})
    # Mock out the relevant methods to test
    monkeypatch.setattr(instrument, "parse_toml_file", MagicMock())
    monkeypatch.setattr(instrument, "connect", AsyncMock(return_value=([], [])))
    monkeypatch.setenv("HAVEN_CONFIG_FILES", str(toml_file), prepend=False)
    # Execute the loading step
    await instrument.load()
    # Check that the right methods were called
    instrument.parse_toml_file.assert_called_once()
    instrument.connect.assert_called_once_with(mock=True, return_exceptions=True)



import logging
import re

import numpy as np
from bluesky_queueserver_api import BPlan
from qtpy import QtWidgets
from qtpy.QtCore import QObject, Signal
from qtpy.QtGui import QDoubleValidator
from xraydb.xraydb import XrayDB

from firefly import display
from firefly.plans.util import is_valid_value, time_converter
from haven.energy_ranges import (
    E_step_to_k_step,
    ERange,
    KRange,
    energy_to_wavenumber,
    k_step_to_E_step,
    merge_ranges,
    wavenumber_to_energy,
)

log = logging.getLogger(__name__)

# How much rounding to do when convert energy to k-space
float_accuracy = 4


# The energy resolution will not be better than 0.01 eV at S25
# e.g. 0.01 eV / 4000 eV -> 10^-6 level, Si(311) is 10-5 level
def wavenumber_to_energy_round(wavenumber):
    return round(wavenumber_to_energy(wavenumber), 2)


def k_step_to_E_step_round(k_start, k_step):
    return round(k_step_to_E_step(k_start, k_step), 2)


class TitleRegion:
    def __init__(self):
        self.setup_ui()

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # Enable checkbox
        self.regions_all_checkbox = QtWidgets.QCheckBox()
        self.regions_all_checkbox.setChecked(True)
        self.regions_all_checkbox.setFixedWidth(15)

        self.layout.addWidget(self.regions_all_checkbox)

        labels = ["Start", "Stop", "Step", "Weight", "", "Exposure [s]"]
        Qlabels_all = {}
        for label_i in labels:
            Qlabel_i = QtWidgets.QLabel(label_i)
            self.layout.addWidget(Qlabel_i)
            Qlabels_all[label_i] = Qlabel_i

        # fix widths so the labels are aligned with XafsRegions
        Qlabels_all["Weight"].setFixedWidth(57)
        Qlabels_all[""].setFixedWidth(57)
        Qlabels_all["Exposure [s]"].setFixedWidth(68)


class XafsScanRegion(QObject):
    time_calculation_signal = Signal()

    def __init__(self):
        super().__init__()
        self.setup_ui()
        self.kErange = None
        self.xafs_region_time = (
            0  # flag for whether time is calculated correctly, if not, will set to -1
        )

        # List of widgets and their signals to connect to update_total_time
        widgets_signals = [
            (self.start_line_edit, "textChanged"),
            (self.stop_line_edit, "textChanged"),
            (self.step_line_edit, "textChanged"),
            (self.weight_spinbox, "valueChanged"),
            (self.exposure_time_spinbox, "valueChanged"),
            (self.k_space_checkbox, "stateChanged"),
        ]

        # Connect all signals to the update_total_time method
        for widget, signal_name in widgets_signals:
            getattr(widget, signal_name).connect(self.update_total_time)

    def setup_ui(self):
        self.layout = QtWidgets.QHBoxLayout()

        # Enable checkbox
        self.region_checkbox = QtWidgets.QCheckBox()
        self.region_checkbox.setChecked(True)
        self.layout.addWidget(self.region_checkbox)

        # First energy box
        self.start_line_edit = QtWidgets.QLineEdit()
        self.start_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.start_line_edit.setPlaceholderText("Start…")
        self.layout.addWidget(self.start_line_edit)

        # Last energy box
        self.stop_line_edit = QtWidgets.QLineEdit()
        self.stop_line_edit.setValidator(QDoubleValidator())  # only takes floats
        self.stop_line_edit.setPlaceholderText("Stop…")
        self.layout.addWidget(self.stop_line_edit)

        # Energy step box
        self.step_line_edit = QtWidgets.QLineEdit()
        self.step_line_edit.setValidator(
            QDoubleValidator(0.0, float("inf"), 2)  # the step is always bigger than 0
        )
        # only takes positive floats
        self.step_line_edit.setPlaceholderText("Step…")
        self.layout.addWidget(self.step_line_edit)

        # Weight factor box
        self.weight_spinbox = QtWidgets.QDoubleSpinBox()
        self.layout.addWidget(self.weight_spinbox)
        self.weight_spinbox.setDecimals(1)
        self.weight_spinbox.setEnabled(False)

        # K-space checkbox
        self.k_space_checkbox = QtWidgets.QCheckBox()
        self.k_space_checkbox.setText("k-space")
        self.k_space_checkbox.setEnabled(True)
        self.layout.addWidget(self.k_space_checkbox)

        # Exposure time double spin box
        self.exposure_time_spinbox = QtWidgets.QDoubleSpinBox()
        self.exposure_time_spinbox.setValue(1)
        self.layout.addWidget(self.exposure_time_spinbox)

        # Connect the k-space enabled checkbox to the relevant signals
        self.k_space_checkbox.stateChanged.connect(self.update_wavenumber_energy)

    def update_line_edit_value(self, line_edit, conversion_func):
        text = line_edit.text()
        if text:
            converted_value = conversion_func(round(float(text), float_accuracy))
            line_edit.setText(f"{converted_value:.6g}")

    def update_wavenumber_energy(self, is_k_checked):
        # disable weight box when k is not selected
        self.weight_spinbox.setEnabled(is_k_checked)

        # Define conversion functions
        conversion_funcs = {
            self.step_line_edit: (
                E_step_to_k_step
                if is_k_checked
                else lambda x, y: k_step_to_E_step_round(x, y)
            ),
            self.start_line_edit: (
                energy_to_wavenumber if is_k_checked else wavenumber_to_energy_round
            ),
            self.stop_line_edit: (
                energy_to_wavenumber if is_k_checked else wavenumber_to_energy_round
            ),
        }

        # Iterate over line edits and apply corresponding conversion
        for line_edit, func in conversion_funcs.items():
            # Special handling for step_line_edit due to different parameters needed
            if line_edit == self.step_line_edit:
                start_text = self.start_line_edit.text()
                if start_text and line_edit.text():
                    start = float(start_text)
                    step = float(line_edit.text())
                    new_values = (
                        func(start, step) if is_k_checked else func(start, step)
                    )
                    line_edit.setText(f"{new_values:.4g}")
            else:
                self.update_line_edit_value(line_edit, func)

    def update_total_time(self):
        weight = self.weight_spinbox.value()
        exposure_time = self.exposure_time_spinbox.value()

        # prevent invalid inputs such as nan
        try:
            start = round(float(self.start_line_edit.text()), float_accuracy)
            stop = round(float(self.stop_line_edit.text()), float_accuracy)
            step = round(float(self.step_line_edit.text()), float_accuracy)

        # when the round doesn't work for nan values
        except ValueError:
            self.kErange = []
            start, stop, step = float("nan"), float("nan"), float("nan")

        if self.k_space_checkbox.isChecked():
            self.kErange = KRange(
                k_min=start,
                k_max=stop,
                k_step=step,
                k_weight=weight,
                exposure=exposure_time,
            )

        else:
            self.kErange = ERange(
                E_min=start,
                E_max=stop,
                E_step=step,
                weight=weight,
                exposure=exposure_time,
            )

        # Emit the signal regardless of success or failure
        self.time_calculation_signal.emit()


class XafsScanDisplay(display.FireflyDisplay):
    min_energy = 4000
    max_energy = 33000

    def customize_ui(self):
        # Remove the defaut XAFS layout from .ui file
        self.ui.clearLayout(self.ui.region_layout)

        # add title layout
        self.title_region = TitleRegion()
        self.ui.title_layout.addLayout(self.title_region.layout)

        self.reset_default_regions()
        # add absorption edges from XrayDB
        self.xraydb = XrayDB()

        combo_box = self.ui.edge_combo_box
        ltab = self.xraydb.tables["xray_levels"]
        edges = self.xraydb.query(ltab)
        edges = edges.filter(
            ltab.c.absorption_edge < self.max_energy,
            ltab.c.absorption_edge > self.min_energy,
        )
        items = [
            f"{r.element} {r.iupac_symbol} ({int(r.absorption_edge)} eV)"
            for r in edges.all()
        ]
        combo_box.addItems(["Select edge…", *items])

        # Connect the E0 checkbox to the E0 combobox
        self.ui.use_edge_checkbox.stateChanged.connect(self.use_edge)

        # disable the line edits in spin box
        self.ui.regions_spin_box.lineEdit().setReadOnly(True)

        # when regions number changed
        self.ui.regions_spin_box.valueChanged.connect(self.update_regions)
        self.ui.regions_spin_box.editingFinished.connect(self.update_regions)

        # reset button
        self.ui.reset_button.clicked.connect(self.reset_default_regions)

        # Run scan button
        self.ui.run_button.clicked.connect(self.queue_plan)

        # connect checkboxes with all regions' check box
        self.title_region.regions_all_checkbox.stateChanged.connect(
            self.on_regions_all_checkbox
        )
        # connect is_standard with a warning box
        self.ui.checkBox_is_standard.clicked.connect(self.on_is_standard)

        # connect num. of scans with total_time
        self.ui.spinBox_repeat_scan_num.valueChanged.connect(self.update_total_time)

    async def update_devices(self, registry):
        """Set available components in the device list."""
        await super().update_devices(registry)
        await self.ui.detectors_list.update_devices(registry)

    def on_region_checkbox(self):
        for region_i in self.regions:
            is_region_i_checked = region_i.region_checkbox.isChecked()
            region_i.start_line_edit.setEnabled(is_region_i_checked)
            region_i.stop_line_edit.setEnabled(is_region_i_checked)
            region_i.step_line_edit.setEnabled(is_region_i_checked)
            region_i.exposure_time_spinbox.setEnabled(is_region_i_checked)
            region_i.weight_spinbox.setEnabled(is_region_i_checked)
            if not self.use_edge_checkbox.isChecked():
                region_i.k_space_checkbox.setEnabled(False)
            else:
                region_i.k_space_checkbox.setEnabled(is_region_i_checked)

    def on_regions_all_checkbox(self, is_checked):
        for region_i in self.regions:
            region_i.region_checkbox.setChecked(is_checked)

    def use_edge(self, is_checked):
        self.edge_combo_box.setEnabled(is_checked)

        # extract edge values
        match = re.findall(r"\d+\.?\d*", self.edge_combo_box.currentText())
        self.edge_value = round(float(match[-1]), float_accuracy) if match else 0

        # iterate through selected regions
        checked_regions = [
            region_i
            for region_i in self.regions
            if region_i.region_checkbox.isChecked()
        ]
        for region_i in checked_regions:
            # Adjust checkbox based on use_edge_checkbox state
            region_i.k_space_checkbox.setEnabled(is_checked)
            # uncheck k space to convert back to energy values from k values
            region_i.k_space_checkbox.setChecked(False)

            # Convert between absolute energies and relative energies
            for line_edit in [region_i.start_line_edit, region_i.stop_line_edit]:
                text = line_edit.text()
                if text:
                    value = (
                        round(float(text), float_accuracy) - self.edge_value
                        if is_checked
                        else round(float(text), float_accuracy) + self.edge_value
                    )
                    line_edit.setText(f"{value:.6g}")

    def clearLayout(self, layout):
        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                if item.widget():
                    item.widget().deleteLater()

    def reset_default_regions(self):
        default_num_regions = 3

        if hasattr(self, "regions"):
            self.remove_regions(len(self.regions))
        self.regions = []
        self.add_regions(default_num_regions)
        self.ui.regions_spin_box.setValue(default_num_regions)

        # set default values for testing, to be deleted in the future
        pre_edge = [-50, -20, 1]
        XANES_region = [-20, 50, 0.2]
        EXAFS_region = [50, 500, 0.5]
        default_regions = [pre_edge, XANES_region, EXAFS_region]
        for i, region_i in enumerate(self.regions):
            region_i.start_line_edit.setText(str(default_regions[i][0]))
            region_i.stop_line_edit.setText(str(default_regions[i][1]))
            region_i.step_line_edit.setText(str(default_regions[i][2]))

        # reset scan repeat num to 1
        self.ui.spinBox_repeat_scan_num.setValue(1)

    def add_regions(self, num=1):
        for i in range(num):
            region = XafsScanRegion()
            self.ui.regions_layout.addLayout(region.layout)
            self.regions.append(region)
            # disable/enabale regions when selected
            region.region_checkbox.stateChanged.connect(self.on_region_checkbox)
            region.region_checkbox.stateChanged.connect(self.update_total_time)
            # receive time signals from XafsRegion
            region.time_calculation_signal.connect(self.update_total_time)

    def remove_regions(self, num=1):
        for i in range(num):
            layout = self.regions[-1].layout
            # iterate/wait, and delete all widgets in the layout in the end
            while layout.count():
                item = layout.takeAt(0)
                if item.widget():
                    item.widget().deleteLater()
            self.regions.pop()

    def update_regions(self):
        new_region_num = self.ui.regions_spin_box.value()
        old_region_num = len(self.regions)
        diff_region_num = new_region_num - old_region_num

        if diff_region_num < 0:
            self.remove_regions(abs(diff_region_num))
        elif diff_region_num > 0:
            self.add_regions(diff_region_num)

    def update_total_time(self):
        # Summing total_time for all checked regions directly within the sum function using a generator expression
        kEranges_all = [
            region_i.kErange
            for region_i in self.regions
            if region_i.region_checkbox.isChecked()
        ]

        # prevent end points are smaller than start points
        try:
            _, exposures = merge_ranges(*kEranges_all, sort=True)
            total_time_per_scan = exposures.sum()
        except ValueError:
            total_time_per_scan = float("nan")

        # calculate time for each scan
        hr, min, sec = time_converter(total_time_per_scan)
        self.ui.label_hour_scan.setText(str(hr))
        self.ui.label_min_scan.setText(str(min))
        self.ui.label_sec_scan.setText(str(sec))

        # calculate time for entire planf
        num_scan_repeat = self.ui.spinBox_repeat_scan_num.value()
        total_time = num_scan_repeat * total_time_per_scan
        hr_total, min_total, sec_total = time_converter(total_time)

        self.ui.label_hour_total.setText(str(hr_total))
        self.ui.label_min_total.setText(str(min_total))
        self.ui.label_sec_total.setText(str(sec_total))

    def on_is_standard(self, is_checked):
        # if is_standard checked, warn that the data will be used for public
        if is_checked:
            response = QtWidgets.QMessageBox.warning(
                self,
                "Notice",
                "When checking this option, this data will be used by public.",
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )
            if response != QtWidgets.QMessageBox.Yes:
                self.ui.checkBox_is_standard.setChecked(False)

    def queue_plan(self, *args, **kwargs):
        """Execute this plan on the queueserver."""
        # Get parameters from each rows of line regions:
        energy_ranges_all = []

        # iterate through only selected regions
        checked_regions = [
            region_i
            for region_i in self.regions
            if region_i.region_checkbox.isChecked()
        ]
        for region_i in checked_regions:
            try:
                start = round(float(region_i.start_line_edit.text()), float_accuracy)
                stop = round(float(region_i.stop_line_edit.text()), float_accuracy)
                step = round(float(region_i.step_line_edit.text()), float_accuracy)
                weight = region_i.weight_spinbox.value()
                exposure_time = region_i.exposure_time_spinbox.value()
            except ValueError:
                QtWidgets.QMessageBox.warning(
                    self, "Value Error", "Invalid value detected!"
                )
                return None
            if region_i.k_space_checkbox.isChecked():
                energy_ranges_all.append(
                    KRange(
                        k_min=start,
                        k_max=stop,
                        k_step=step,
                        k_weight=weight,
                        exposure=exposure_time,
                    )
                )

            else:
                energy_ranges_all.append(
                    ERange(
                        E_min=start,
                        E_max=stop,
                        E_step=step,
                        weight=weight,
                        exposure=exposure_time,
                    )
                )

        # Turn ndarrays into lists so they can be JSON serialized
        energies, exposures = merge_ranges(*energy_ranges_all, sort=True)
        energies = list(np.round(energies, float_accuracy))
        exposures = list(np.round(exposures, float_accuracy))
        detectors = self.ui.detectors_list.selected_detectors()
        repeat_scan_num = int(self.ui.spinBox_repeat_scan_num.value())
        md = {
            "sample": self.ui.lineEdit_sample.text(),
            "purpose": self.ui.lineEdit_purpose.text(),
            "is_standard": self.ui.checkBox_is_standard.isChecked(),
            "notes": self.ui.textEdit_notes.toPlainText(),
        }
        # Only include metadata that isn't an empty string
        md = {key: val for key, val in md.items() if is_valid_value(val)}

        # Check that an absorption edge was selected
        if self.use_edge_checkbox.isChecked():
            try:
                match = re.findall(r"\d+\.?\d*", self.edge_combo_box.currentText())
                self.edge_value = round(float(match[-1]), float_accuracy)

            except IndexError:
                QtWidgets.QMessageBox.warning(
                    self, "Error", "Please select an absorption edge."
                )
                return None
        else:
            self.edge_value = 0

        # Build the queue item
        item = BPlan(
            "energy_scan",
            energies=energies,
            exposure=exposures,
            E0=self.edge_value,
            detectors=detectors,
            md=md,
        )
        # Submit the item to the queueserver
        log.info(f"Adding XAFS scan to queue.")
        # repeat scans
        for i in range(repeat_scan_num):
            self.queue_item_submitted.emit(item)

    def ui_filename(self):
        return "plans/xafs_scan.ui"


# -----------------------------------------------------------------------------
# :author:    Juanjuan Huang
# :email:     juanjuan.huang@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from firefly.iocs import IocsDisplay

dummy_config = {"iocs"}


def test_embedded_display_widgets(qtbot, beamline_manager, sim_registry):
    """Test the the voltmeters creates a new embedded display widget for
    each ion chamber.

    """
    # Load the display
    display = IocsDisplay()
    qtbot.addWidget(display)
    # Check that the embedded display widgets get added correctly
    assert hasattr(display, "_ioc_displays")
    assert len(display._ioc_displays) == 2
    # two displays and a separator
    assert display.iocs_layout.count() == 3
    # # Check that the embedded display widgets have the correct macros
    emb_disp = display._ioc_displays[0]
    disp = emb_disp.open_file(force=True)
    macros = disp.macros()
    assert macros == {"NAME": "255idb", "IOC": "companion_cube_iocs_ioc255idb"}



"""LabJack Data Acquisition (DAQ)
+++++++++++++++++++++++++++++++++++++++

Ophyd-async definitions for Labjack T-series data acquisition devices.

Supported devices, all inherit from ``LabJackBase``:

- T4
- T7
- T7Pro
- T8

These devices are **based on EPICS LabJack module R3.0**. The EPICS IOC
database changed significantly from R2 to R3 when the module was
rewritten to use the LJM library.

.. seealso:: https://github.com/epics-modules/LabJack/releases/tag/R3-0

There are definitions for the entire LabJack device, as well as the
various inputs/outputs available on the LabJack T-series.
Individual inputs can be used as part of other devices. Assuming
analog input 5 is connected to a gas flow meter:

.. code:: python

    from ophyd_async.core import StandardReadable
    from haven.instrument import labjack

    class MyBeamline(StandardReadable):
        def __init__(self, prefix: str, name: str = ""):
            self.gas_flow = labjack.AnalogInput("LabJackT7_1:Ai5")
            ...   # Other devices
            super().__init__(name=name)

"""

import asyncio

import numpy as np
from bluesky.protocols import Triggerable
from numpy.typing import NDArray
from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    ConfigSignal,
    DeviceVector,
    HintedSignal,
    StandardReadable,
    SubsetEnum,
    observe_value,
)
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw, epics_signal_x

from ..typing import StrEnum
from .synApps import EpicsRecordInputFields, EpicsRecordOutputFields

__all__ = [
    "AnalogOutput",
    "AnalogInput",
    "DigitalIO",
    "WaveformDigitizer",
    "WaveformGenerator",
    "LabJackBase",
    "LabJackT4",
    "LabJackT7",
    "LabJackT7Pro",
    "LabJackT8",
]

KIND_CONFIG_OR_NORMAL = 3
"""Alternative for ``Kind.config | Kind.normal``."""


class Input(EpicsRecordInputFields):
    """A generic input record.

    Similar to synApps input records but with some changes. The .PROC
    field is used as a trigger. This way, even if the .SCAN field is
    set to passive, the record can still be updated before being read.
    """

    def __init__(self, prefix: str, name: str = ""):
        self.process_record = epics_signal_x(f"{prefix}.PROC")
        super().__init__(prefix=prefix, name=name)


class BinaryInput(Input):

    DeviceType = SubsetEnum[
        "Soft Channel",
        "Raw Soft Channel",
        "Async Soft Channel",
        "Db State",
        "asynInt32",
        "asynUInt32Digital",
        "asyn bi stringParm",
        "asyn MPC",
        "Vg307 GPIB Instrument",
        "asyn Televac",
        "asyn TPG261",
        "lua",
        "stream",
        "EtherIP",
        "dg535",
        "Zed",
    ]

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables():
            self.final_value = epics_signal_r(bool, f"{prefix}.VAL")
        self.raw_value = epics_signal_rw(float, f"{prefix}.RVAL")
        super().__init__(prefix=prefix, name=name)


class Output(EpicsRecordOutputFields):
    """A generic output record.

    Intended to be sub-classed into different output types.

    """


class BinaryOutput(Output):
    """A binary input on the labjack."""

    DeviceType = SubsetEnum[
        "Soft Channel",
        "Raw Soft Channel",
        "Async Soft Channel",
        "General Time",
        "Db State",
        "asynInt32",
        "asynUInt32Digital",
        "asyn bo stringParm",
        "asyn MPC",
        "Vg307 GPIB Instrument",
        "PZT Bug",
        "asyn TPG261",
        "lua",
        "stream",
        "EtherIP",
        "dg535",
    ]

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables():
            self.desired_value = epics_signal_rw(bool, f"{prefix}.VAL")
        self.raw_value = epics_signal_rw(float, f"{prefix}.RVAL")
        self.readback_value = epics_signal_r(float, f"{prefix}.RBV")


class AnalogOutput(Output):
    """An analog output on a labjack device."""

    DeviceType = SubsetEnum[
        "Soft Channel",
        "Raw Soft Channel",
        "Async Soft Channel",
        "asynInt32",
        "asynFloat64",
        "asynInt64",
        "IOC stats",
        "asyn ao stringParm",
        "asyn ao Eurotherm",
        "asyn MPC",
        "PZT Bug",
        "asyn TPG261",
        "lua",
        "stream",
        "EtherIP",
        "dg535",
    ]

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables():
            self.desired_value = epics_signal_rw(float, f"{prefix}.VAL")
        self.raw_value = epics_signal_rw(int, f"{prefix}.RVAL")
        with self.add_children_as_readables(HintedSignal):
            self.readback_value = epics_signal_r(int, f"{prefix}.RBV")
        super().__init__(prefix=prefix, name=name)


class AnalogInput(Input, Triggerable):
    """An analog input on a labjack device.

    It is based on the synApps input record, but with LabJack specific
    signals added.

    """

    class DifferentialMode(StrEnum):
        SINGLE_ENDED = "Single-Ended"
        DIFFERENTIAL = "Differential"

    class TemperatureUnits(StrEnum):
        KELVIN = "K"
        CELSIUS = "C"
        FAHRENHEIT = "F"

    class Mode(StrEnum):
        VOLTS = "Volts"
        TYPE_B_TC = "Type B TC"
        TYPE_C_TC = "Type C TC"
        TYPE_E_TC = "Type E TC"
        TYPE_J_TC = "Type J TC"
        TYPE_K_TC = "Type K TC"
        TYPE_N_TC = "Type N TC"
        TYPE_R_TC = "Type R TC"
        TYPE_S_TC = "Type S TC"
        TYPE_T_TC = "Type T TC"

    class Range(StrEnum):
        TEN_VOLTS = "+= 10V"
        ONE_VOLT = "+= 1V"
        HUNDRED_MILLIVOLTS = "+= 0.1V"
        TEN_MILLIVOLTS = "+= 0.01V"

    class Resolution(StrEnum):
        DEFAULT = "Default"
        ONE = "1"
        TWO = "2"
        THREE = "3"
        FOUR = "4"
        FIVE = "5"
        SIX = "6"
        SEVEN = "7"
        EIGHT = "8"

    DeviceType = SubsetEnum[
        "Soft Channel",
        "Raw Soft Channel",
        "Async Soft Channel",
        "Soft Timestamp",
        "General Time",
        "asynInt32",
        "asynInt32Average",
        "asynFloat64",
        "asynFloat64Average",
        "asynInt64",
        "IOC stats",
        "IOC stats clusts",
        "GPIB init/report",
        "Sec Past Epoch",
        "asyn ai stringParm",
        "asyn ai HeidND261",
    ]

    def __init__(self, prefix: str, ch_num: int, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.differential = epics_signal_rw(
                self.DifferentialMode, f"{prefix}AiDiff{ch_num}"
            )
            self.high = epics_signal_rw(float, f"{prefix}AiHOPR{ch_num}")
            self.low = epics_signal_rw(float, f"{prefix}AiLOPR{ch_num}")
            self.temperature_units = epics_signal_rw(
                self.TemperatureUnits, f"{prefix}AiTempUnits{ch_num}"
            )

            self.resolution = epics_signal_rw(
                self.Resolution, f"{prefix}AiResolution{ch_num}"
            )
            self.range = epics_signal_rw(self.Range, f"{prefix}AiRange{ch_num}")
            self.mode = epics_signal_rw(self.Mode, f"{prefix}AiMode{ch_num}")
            self.enable = epics_signal_rw(bool, f"{prefix}AiEnable{ch_num}")
        with self.add_children_as_readables(HintedSignal):
            self.final_value = epics_signal_r(float, f"{prefix}Ai{ch_num}.VAL")
        self.raw_value = epics_signal_rw(int, f"{prefix}Ai{ch_num}.RVAL")
        super().__init__(prefix=f"{prefix}Ai{ch_num}", name=name)

    @AsyncStatus.wrap
    async def trigger(self):
        await self.process_record.trigger()


class DigitalIO(StandardReadable):
    """A digital input/output channel on the labjack.

    Because of the how the records are structured in EPICS, the prefix
    must not include the "Bi{N}" portion of the prefix. Instead, the
    prefix should be prefix for the whole labjack
    (e.g. ``LabJackT7_1:``), and the channel number should be provided
    using the *ch_num* property. So for the digital I/O with its input
    available at PV ``LabJackT7_1:Bi3``, use:

    .. code:: python

        dio3 = DigitalIO("LabJackT7_1:", name="dio3", ch_num=3)

    This will create signals for the input (``Bi3``), output
    (``Bo3``), and direction (``Bd3``) records.

    """

    class Direction(StrEnum):
        INPUT = "In"
        OUTPUT = "Out"

    def __init__(self, prefix: str, ch_num: int, name: str = ""):
        with self.add_children_as_readables():
            self.input = BinaryInput(f"{prefix}Bi{ch_num}")
            self.output = BinaryOutput(f"{prefix}Bo{ch_num}")
        with self.add_children_as_readables(ConfigSignal):
            self.direction = epics_signal_rw(self.Direction, f"{prefix}Bd{ch_num}")
        super().__init__(name=name)


class WaveformDigitizer(StandardReadable, Triggerable):
    """A feature of the Labjack devices that allows waveform capture."""

    class TriggerSource(StrEnum):
        INTERNAL = "Internal"
        EXTERNAL = "External"

    def __init__(self, prefix: str, name: str = "", waveforms=[]):
        with self.add_children_as_readables():
            self.timebase_waveform = epics_signal_rw(
                NDArray[np.float64], f"{prefix}WaveDigTimeWF"
            )
            self.dwell_actual = epics_signal_rw(float, f"{prefix}WaveDigDwellActual")
            self.total_time = epics_signal_rw(float, f"{prefix}WaveDigTotalTime")
        with self.add_children_as_readables(ConfigSignal):
            self.num_points = epics_signal_rw(int, f"{prefix}WaveDigNumPoints")
            self.dwell_time = epics_signal_rw(float, f"{prefix}WaveDigDwell")
            self.first_chan = epics_signal_rw(
                SubsetEnum[
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                ],
                f"{prefix}WaveDigFirstChan",
            )
            self.num_chans = epics_signal_rw(
                SubsetEnum[
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "10",
                    "11",
                    "12",
                    "13",
                    "14",
                ],
                f"{prefix}WaveDigNumChans",
            )
            self.resolution = epics_signal_rw(
                SubsetEnum["Default", "1", "2", "3", "4", "5", "6", "7", "8"],
                f"{prefix}WaveDigResolution",
            )
            self.settling_time = epics_signal_rw(float, f"{prefix}WaveDigSettlingTime")
        self.current_point = epics_signal_rw(int, f"{prefix}WaveDigCurrentPoint")
        self.ext_trigger = epics_signal_rw(
            self.TriggerSource, f"{prefix}WaveDigExtTrigger"
        )
        self.ext_clock = epics_signal_rw(self.TriggerSource, f"{prefix}WaveDigExtClock")
        self.auto_restart = epics_signal_x(f"{prefix}WaveDigAutoRestart")
        self.run = epics_signal_rw(bool, f"{prefix}WaveDigRun")
        self.read_waveform = epics_signal_rw(
            SubsetEnum["Done", "Read"], f"{prefix}WaveDigReadWF"
        )
        # Add waveforms
        with self.add_children_as_readables():
            self.waveforms = DeviceVector(
                {
                    idx: epics_signal_r(
                        NDArray[np.float64], f"{prefix}WaveDigVoltWF{idx}"
                    )
                    for idx in waveforms
                }
            )
        super().__init__(name=name)

    @AsyncStatus.wrap
    async def trigger(self):
        # Determine how long the trigger will take
        dwell_time, num_points = await asyncio.gather(
            self.dwell_time.get_value(),
            self.num_points.get_value(),
        )
        timeout = dwell_time * num_points + DEFAULT_TIMEOUT
        # Start the triggering
        await self.run.set(True, wait=False)
        # Wait for the read to complete
        async for value in observe_value(self.run, timeout=timeout):
            if not value:
                break


class WaveformGenerator(StandardReadable):
    """A feature of the Labjack devices that generates output waveforms."""

    class WaveType(StrEnum):
        USER_DEFINED = "User-defined"
        SINE_WAVE = "Sin wave"
        SQUARE_WAVE = "Square wave"
        SAWTOOTH = "Sawtooth"
        PULSE = "Pulse"
        RANDOM = "Random"

    TriggerSource = WaveformDigitizer.TriggerSource

    class TriggerMode(StrEnum):
        ONE_SHOT = "One-shot"
        CONTINUOS = "Continuous"

    def __init__(self, prefix: str, name: str = ""):
        with self.add_children_as_readables(ConfigSignal):
            self.external_trigger = epics_signal_rw(
                self.TriggerSource, f"{prefix}WaveGenExtTrigger"
            )
            self.external_clock = epics_signal_rw(
                self.TriggerSource, f"{prefix}WaveGenExtClock"
            )
            self.continuous = epics_signal_rw(
                self.TriggerMode, f"{prefix}WaveGenContinuous"
            )
        self.run = epics_signal_x(f"{prefix}WaveGenRun")

        # These signals give a readback based on whether user-defined or
        # internal waves are used
        with self.add_children_as_readables():
            self.frequency = epics_signal_r(float, f"{prefix}WaveGenFrequency")
            self.dwell = epics_signal_r(float, f"{prefix}WaveGenDwell")
            self.dwell_actual = epics_signal_r(float, f"{prefix}WaveGenDwellActual")
            self.total_time = epics_signal_r(float, f"{prefix}WaveGenTotalTime")
        with self.add_children_as_readables(ConfigSignal):
            self.num_points = epics_signal_r(int, f"{prefix}WaveGenNumPoints")
        self.current_point = epics_signal_r(int, f"{prefix}WaveGenCurrentPoint")

        # Settings for user-defined waveforms
        with self.add_children_as_readables():
            self.user_time_waveform = epics_signal_rw(
                NDArray[np.float64], f"{prefix}WaveGenUserTimeWF"
            )
        self.user_num_points = epics_signal_rw(int, f"{prefix}WaveGenUserNumPoints")
        self.user_dwell = epics_signal_rw(float, f"{prefix}WaveGenUserDwell")
        self.user_frequency = epics_signal_rw(float, f"{prefix}WaveGenUserFrequency")

        # Settings for internal waveforms
        with self.add_children_as_readables():
            self.internal_time_waveform = epics_signal_rw(
                NDArray[np.float64], f"{prefix}WaveGenIntTimeWF"
            )
        self.internal_num_points = epics_signal_rw(int, f"{prefix}WaveGenIntNumPoints")
        self.internal_dwell = epics_signal_rw(float, f"{prefix}WaveGenIntDwell")
        self.internal_frequency = epics_signal_rw(float, f"{prefix}WaveGenIntFrequency")

        # Waveform specific settings
        with self.add_children_as_readables(ConfigSignal):
            self.user_waveform_0 = epics_signal_rw(
                NDArray[np.float64], f"{prefix}WaveGenUserWF0"
            )
            self.enable_0 = epics_signal_rw(bool, f"{prefix}WaveGenEnable0")
            self.type_0 = epics_signal_rw(self.WaveType, f"{prefix}WaveGenType0")
            self.pulse_width_0 = epics_signal_rw(float, f"{prefix}WaveGenPulseWidth0")
            self.amplitude_0 = epics_signal_rw(float, f"{prefix}WaveGenAmplitude0")
            self.offset_0 = epics_signal_rw(float, f"{prefix}WaveGenOffset0")
            self.user_waveform_1 = epics_signal_rw(
                NDArray[np.float64], f"{prefix}WaveGenUserWF1"
            )
            self.enable_1 = epics_signal_rw(bool, f"{prefix}WaveGenEnable1")
            self.type_1 = epics_signal_rw(self.WaveType, f"{prefix}WaveGenType1")
            self.pulse_width_1 = epics_signal_rw(float, f"{prefix}WaveGenPulseWidth1")
            self.amplitude_1 = epics_signal_rw(float, f"{prefix}WaveGenAmplitude1")
            self.offset_1 = epics_signal_rw(float, f"{prefix}WaveGenOffset1")
        self.internal_waveform_0 = epics_signal_rw(
            NDArray[np.float64], f"{prefix}WaveGenInternalWF0"
        )
        self.internal_waveform_1 = epics_signal_r(
            NDArray[np.float64], f"{prefix}WaveGenInternalWF1"
        )

        super().__init__(name=name)


class LabJackBase(StandardReadable):
    """A labjack T-series data acquisition unit (DAQ).

    To use the individual components separately, consider using the
    corresponding devices in the list below.

    This device contains signals for the following:

    - device information (e.g. firmware version ,etc)
    - analog outputs (:py:class:`~haven.instrument.labjack.AnalogInput`)
    - analog inputs* (:py:class:`~haven.instrument.labjack.AnalogOutput`)
    - digital input/output* (:py:class:`~haven.instrument.labjack.DigitalIO`)
    - waveform digitizer* (:py:class:`~haven.instrument.labjack.WaveformDigitizer`)
    - waveform generator (:py:class:`~haven.instrument.labjack.WaveformGenerator`)

    The number of inputs and digital outputs depends on the specific
    LabJack T-series device being used. Therefore, the base device
    ``LabJackBase`` can create arbitrary configurations of these I/O
    signals. For a specific model, consider using one of the
    subclasses, like ``LabJackT4``.

    The waveform generator and waveform digitizer are included for
    convenience. Reading all the analog/digital inputs and outputs can
    be done by calling the ``.read()`` method. However, it is unlikely
    that the goal is also to trigger the digitizer and generator
    during this read. For this reason, **the digitizer and generator
    are not included as readables**. To read or trigger the digitizer or
    generator, they must be used as separate devices:

    .. code:: python

        lj = LabJackT4(...)

        # Read a waveform from the digitizer
        await lj.waveform_digitizer.trigger()
        await lj.waveform_digitizer.read()

        # Same thing for the waveform generator
        await lj.waveform_generator.trigger()

    """

    Resolution = AnalogInput.Resolution
    DeviceType = SubsetEnum[
        "Soft Channel",
        "Raw Soft Channel",
        "Async Soft Channel",
        "Soft Timestamp",
        "General Time",
        "asynInt32",
        "asynInt32Average",
        "asynFloat64",
        "asynFloat64Average",
        "asynInt64",
        "IOC stats",
        "IOC stats clusts",
        "GPIB init/report",
        "Sec Past Epoch",
        "asyn ai stringParm",
        "asyn ai HeidND261",
    ]

    class Model(StrEnum):
        T4 = "T4"
        T7 = "T7"
        T7_PRO = "T7-Pro"
        T8 = "T8"

    def __init__(
        self,
        prefix: str,
        name: str = "",
        analog_inputs=[],
        digital_ios=[],
        analog_outputs=range(2),
        digital_words=["dio", "eio", "fio", "mio", "cio"],
    ):
        with self.add_children_as_readables(ConfigSignal):
            self.model_name = epics_signal_r(self.Model, f"{prefix}ModelName")
            self.firmware_version = epics_signal_r(str, f"{prefix}FirmwareVersion")
            self.serial_number = epics_signal_r(str, f"{prefix}SerialNumber")
            self.device_temperature = epics_signal_r(
                float, f"{prefix}DeviceTemperature"
            )
            self.ljm_version = epics_signal_r(str, f"{prefix}LJMVersion")
            self.driver_version = epics_signal_r(str, f"{prefix}DriverVersion")
            self.last_error_message = epics_signal_r(
                NDArray[np.uint8], f"{prefix}LastErrorMessage"
            )
            self.poll_sleep_ms = epics_signal_rw(float, f"{prefix}PollSleepMS")
            self.analog_in_settling_time_all = epics_signal_rw(
                float, f"{prefix}AiAllSettlingUS"
            )
            self.analog_in_resolution_all = epics_signal_rw(
                self.Resolution, f"{prefix}AiAllResolution"
            )
            self.analog_in_sampling_rate = epics_signal_rw(
                float,
                write_pv=f"{prefix}AiSamplingRate",
                read_pv=f"{prefix}AiSamplingRate_RBV",
            )
        self.poll_time_ms = epics_signal_r(float, f"{prefix}PollTimeMS")
        self.device_reset = epics_signal_x(f"{prefix}DeviceReset")

        # Common sub-devices (all labjacks have 2 analog outputs)
        # NB: Analog inputs/digital I/Os are on a per-model basis
        with self.add_children_as_readables():
            self.analog_outputs = DeviceVector(
                {idx: AnalogOutput(f"{prefix}Ao{idx}") for idx in analog_outputs}
            )
            self.analog_inputs = DeviceVector(
                {idx: AnalogInput(prefix, ch_num=idx) for idx in analog_inputs}
            )
            # Digital I/O channels
            self.digital_ios = DeviceVector(
                {idx: DigitalIO(prefix, ch_num=idx) for idx in digital_ios}
            )
            self.digital_words = DeviceVector(
                {
                    word: epics_signal_r(int, f"{prefix}{word.upper()}In")
                    for word in digital_words
                }
            )
        # Waveform devices (not read by default, should be made readable as needed)
        self.waveform_digitizer = WaveformDigitizer(
            f"{prefix}", waveforms=analog_inputs
        )
        self.waveform_generator = WaveformGenerator(f"{prefix}")

        super().__init__(name=name)


class LabJackT4(LabJackBase):
    # Inherit the docstring from the base class
    # (needed for sphinx auto API)
    __doc__ = LabJackBase.__doc__

    def __init__(
        self,
        prefix: str,
        name: str = "",
        analog_inputs=range(12),
        digital_ios=range(16),
        analog_outputs=range(2),
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            analog_inputs=analog_inputs,
            digital_ios=digital_ios,
            analog_outputs=analog_outputs,
            **kwargs,
        )


class LabJackT7(LabJackBase):
    # Inherit the docstring from the base class
    # (needed for sphinx auto API)
    __doc__ = LabJackBase.__doc__

    def __init__(
        self,
        prefix: str,
        name: str = "",
        analog_inputs=range(14),
        digital_ios=range(23),
        analog_outputs=range(2),
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            analog_inputs=analog_inputs,
            digital_ios=digital_ios,
            analog_outputs=analog_outputs,
            **kwargs,
        )


class LabJackT7Pro(LabJackBase):
    # Inherit the docstring from the base class
    # (needed for sphinx auto API)
    __doc__ = LabJackBase.__doc__

    def __init__(
        self,
        prefix: str,
        name: str = "",
        analog_inputs=range(14),
        digital_ios=range(23),
        analog_outputs=range(2),
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            analog_inputs=analog_inputs,
            digital_ios=digital_ios,
            analog_outputs=analog_outputs,
            **kwargs,
        )


class LabJackT8(LabJackBase):
    # Inherit the docstring from the base class
    # (needed for sphinx auto API)
    __doc__ = LabJackBase.__doc__

    def __init__(
        self,
        prefix: str,
        name: str = "",
        analog_inputs=range(8),
        digital_ios=range(20),
        analog_outputs=range(2),
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            analog_inputs=analog_inputs,
            digital_ios=digital_ios,
            analog_outputs=analog_outputs,
            **kwargs,
        )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
from caproto.server import PVGroup, ioc_arg_parser, pvproperty, run


class SimpleGroup(PVGroup):
    """
    An IOC with three uncoupled read/writable PVs

    Scalar PVs
    ----------
    A (int)
    B (float)

    Vectors PVs
    -----------
    C (vector of int)
    """

    A = pvproperty(value=1, doc="An integer")
    B = pvproperty(value=2.0, doc="A float")
    C = pvproperty(value=[1, 2, 3], doc="An array of integers")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="simple:", desc="haven.tests.ioc_undulator test IOC"
    )
    ioc = SimpleGroup(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import qtawesome as qta

from firefly import display


class TweakDisplay(display.FireflyDisplay):
    """A small set of widgets for tweaking a signal by specified value."""

    def customize_ui(self):
        # Set button icons
        direction = self.macros().get("DIRECTION", "").lower()
        if direction == "vertical":
            forward_icon = qta.icon("fa5s.arrow-up")
            reverse_icon = qta.icon("fa5s.arrow-down")
        elif direction == "horizontal":
            forward_icon = qta.icon("fa5s.arrow-right")
            reverse_icon = qta.icon("fa5s.arrow-left")
        else:
            forward_icon = qta.icon("fa5s.plus")
            reverse_icon = qta.icon("fa5s.minus")
        self.ui.forward_button.setIcon(forward_icon)
        self.ui.reverse_button.setIcon(reverse_icon)
        for btn in [self.ui.reverse_button, self.ui.forward_button]:
            btn.setText("")
        if hasattr(self.ui, "value_spin_box"):
            self.ui.value_spin_box.valueChanged.connect(self.update_press_values)

    def update_press_values(self, tweak_value):
        """Update the press values for the buttons.

        This makes them move in increments of the tweak value.

        """
        self.ui.reverse_button.pressValue = -tweak_value
        self.ui.forward_button.pressValue = tweak_value

    def ui_filename(self):
        return "tweak.ui"



from ophyd.sim import motor1, motor2, motor3

from haven import robot_transfer_sample


def test_robot_sample(robot):
    plan = robot_transfer_sample(robot, 9, motor1, 100, motor2, 200, motor3, 50)

    msgs = list(plan)

    assert robot.name == "robotA"
    assert len(msgs) == 14

    unload = robot_transfer_sample(robot, None, motor1, 100)
    msgs = list(unload)
    assert len(msgs) == 6


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
import logging
import sqlite3
import threading
import warnings
from functools import partial

import databroker
import numpy as np
import pandas as pd
from tiled.client import from_uri
from tiled.client.cache import Cache

from ._iconfig import load_config

log = logging.getLogger(__name__)


def unsnake(arr: np.ndarray, snaking: list) -> np.ndarray:
    """Unsnake a nump array.

    For each axis in *arr*, there should be a corresponding True/False
    in *snaking* whether that axis should have alternating rows. The
    first entry is ignored as it doesn't make sense to snake the first
    axis.

    Returns
    =======
    unsnaked
      A copy of *arr* with the odd-numbered axes flipped (if indicated
      by *snaking*).

    """
    # arr = np.copy(arr)
    # Create some slice object for easier manipulation
    full_axis = slice(None)
    alternating = slice(None, None, 2)
    flipped = slice(None, None, -1)
    # Flip each axis if necessary (skipping the first axis)
    for axis, is_snaked in enumerate(snaking[1:]):
        if not is_snaked:
            continue
        slices = (full_axis,) * axis
        slices += (alternating,)
        arr[slices] = arr[slices + (flipped,)]
    return arr


def load_catalog(name: str = "bluesky"):
    """Load a databroker catalog for retrieving data.

    To retrieve individual scans, consider the ``load_result`` and
    ``load_data`` methods.

    Parameters
    ==========
    name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)

    Returns
    =======
    catalog
      The databroker catalog.
    """
    return databroker.catalog[name]


def load_result(uid: str, catalog_name: str = "bluesky", stream: str = "primary"):
    """Load a past experiment from the database.

    The result contains metadata and scan parameters. The data
    themselves are accessible from the result's *read()* method.

    Parameters
    ==========
    uid
      The universal identifier for this scan, as return by a bluesky
      RunEngine.
    catalog_name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)
    stream
      The data stream defined by the bluesky RunEngine.

    Returns
    =======
    result
      The experiment result, with data available via the *read()*
      method.

    """
    cat = load_catalog(name=catalog_name)
    result = cat[uid][stream]
    return result


def load_data(uid, catalog_name="bluesky", stream="primary"):
    """Load a past experiment's data from the database.

    The result is an xarray with the data collected.

    Parameters
    ==========
    uid
      The universal identifier for this scan, as return by a bluesky
      RunEngine.
    catalog_name
      The name of the catalog as defined in the Intake file
      (e.g. ~/.local/share/intake/catalogs.yml)
    stream
      The data stream defined by the bluesky RunEngine.

    Returns
    =======
    data
      The experimental data, as an xarray.

    """

    res = load_result(uid=uid, catalog_name=catalog_name, stream=stream)
    data = res.read()
    return data


def with_thread_lock(fn):
    """Makes sure the function isn't accessed concurrently."""

    def wrapper(obj, *args, **kwargs):
        obj._lock.acquire()
        try:
            fn(obj, *args, **kwargs)
        finally:
            obj._lock.release()

    return wrapper


class ThreadSafeCache(Cache):
    """Equivalent to the regular cache, but thread-safe.

    Ensures that sqlite3 is built with concurrency features, and
    ensures that no two write operations happen concurrently.

    """

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._lock = threading.Lock()

    def write_safe(self):
        """
        Check that it is safe to write.

        SQLite is not threadsafe for concurrent _writes_.
        """
        is_main_thread = threading.current_thread().ident == self._owner_thread
        sqlite_is_safe = sqlite3.threadsafety == 1
        return is_main_thread or sqlite_is_safe

    # Wrap the accessor methods so they wait for the lock
    clear = with_thread_lock(Cache.clear)
    set = with_thread_lock(Cache.set)
    get = with_thread_lock(Cache.get)
    delete = with_thread_lock(Cache.delete)


def tiled_client(
    entry_node=None, uri=None, cache_filepath=None, structure_clients="dask"
):
    config = load_config()
    # Create a cache for saving local copies
    if cache_filepath is None:
        cache_filepath = config["database"].get("tiled", {}).get("cache_filepath", "")
        cache_filepath = cache_filepath or None
    cache = ThreadSafeCache(filepath=cache_filepath)
    # Create the client
    if uri is None:
        uri = config["database"]["tiled"]["uri"]
    client_ = from_uri(uri, structure_clients)
    if entry_node is None:
        entry_node = config["database"]["tiled"]["entry_node"]
    client_ = client_[entry_node]
    return client_


class CatalogScan:
    """A single scan from the tiled API with some convenience methods.

    Parameters
    ==========
      A tiled container on which to operate."""

    def __init__(self, container):
        self.container = container

    def _read_data(self, signals, dataset="primary/data"):
        # Fetch data if needed
        data = self.container[dataset]
        return data.read(signals)

    def _read_metadata(self, keys=None):
        container = self.container
        if keys is not None:
            container = container[keys]
        return container.metadata

    @property
    def uid(self):
        return self.container._item["id"]

    async def export(self, filename: str, format: str):
        target = partial(self.container.export, filename, format=format)
        await self.loop.run_in_executor(None, target)

    def formats(self):
        return self.container.formats

    async def data(self, stream="primary"):
        return await self.loop.run_in_executor(
            None, self._read_data, None, f"{stream}/data"
        )

    async def to_dataframe(self, signals=None):
        """Convert the dataset into a pandas dataframe."""
        xarray = await self.loop.run_in_executor(None, self._read_data, signals)
        if len(xarray) > 0:
            df = xarray.to_dataframe()
        else:
            df = pd.DataFrame()
        return df

    @property
    def loop(self):
        return asyncio.get_running_loop()

    async def data_keys(self, stream="primary"):
        stream_md = await self.loop.run_in_executor(None, self._read_metadata, stream)
        # Assumes the 0-th descriptor is for the primary stream
        return stream_md["descriptors"][0]["data_keys"]

    async def hints(self):
        """Retrieve the data hints for this scan.

        Returns
        =======
        independent
          The hints for the independent scanning axis.
        dependent
          The hints for the dependent scanning axis.
        """
        metadata = await self.metadata
        # Get hints for the independent (X)
        try:
            independent = metadata["start"]["hints"]["dimensions"][0][0]
        except (KeyError, IndexError):
            warnings.warn("Could not get independent hints")
        # Get hints for the dependent (X)
        dependent = []
        primary_metadata = await self.loop.run_in_executor(
            None, self._read_metadata, "primary"
        )
        hints = primary_metadata["descriptors"][0]["hints"]
        for device, dev_hints in hints.items():
            dependent.extend(dev_hints["fields"])
        return independent, dependent

    @property
    async def metadata(self):
        metadata = await self.loop.run_in_executor(None, self._read_metadata)
        return metadata

    async def __getitem__(self, signal):
        """Retrieve a signal from the dataset, with reshaping etc."""
        loop = asyncio.get_running_loop()
        arr = await loop.run_in_executor(None, self._read_data, tuple([signal]))
        arr = np.asarray(arr[signal])
        # Re-shape to match the scan dimensions
        metadata = await self.metadata
        try:
            shape = metadata["start"]["shape"]
        except KeyError:
            log.warning(f"No shape found for {repr(signal)}.")
        else:
            arr = np.reshape(arr, shape)
        # Flip alternating rows if snaking is enabled
        if "snaking" in metadata["start"]:
            arr = unsnake(arr, metadata["start"]["snaking"])
        return arr


class Catalog:
    """An asynchronous wrapper around the tiled client.

    This class has a more intelligent understanding of how *our* data
    are structured, so can make some assumptions and takes care of
    boiler-plate code (e.g. reshaping maps, etc).

    """

    _client = None

    def __init__(self, client=None):
        self._client = client

    @property
    def loop(self):
        return asyncio.get_running_loop()

    @property
    async def client(self):
        if self._client is None:
            self._client = await self.loop.run_in_executor(None, tiled_client)
        return self._client

    async def __getitem__(self, uid) -> CatalogScan:
        client = await self.client
        container = await self.loop.run_in_executor(None, client.__getitem__, uid)
        scan = CatalogScan(container=container)
        return scan

    async def items(self):
        client = await self.client
        for key, value in await self.loop.run_in_executor(None, client.items):
            yield key, CatalogScan(container=value)

    async def values(self):
        client = await self.client
        containers = await self.loop.run_in_executor(None, client.values)
        for container in containers:
            yield CatalogScan(container)

    async def __len__(self):
        client = await self.client
        length = await self.loop.run_in_executor(None, client.__len__)
        return length

    async def search(self, query):
        """
        Make a Node with a subset of this Node's entries, filtered by query.

        Examples
        --------

        >>> from tiled.queries import FullText
        >>> await tree.search(FullText("hello"))
        """
        loop = asyncio.get_running_loop()
        client = await self.client
        return Catalog(await loop.run_in_executor(None, client.search, query))

    async def distinct(
        self, *metadata_keys, structure_families=False, specs=False, counts=False
    ):
        """Get the unique values and optionally counts of metadata_keys,
        structure_families, and specs in this Node's entries

        Examples
        --------

        Query all the distinct values of a key.

        >>> await catalog.distinct("foo", counts=True)

        Query for multiple keys at once.

        >>> await catalog.distinct("foo", "bar", counts=True)

        """
        loop = asyncio.get_running_loop()
        client = await self.client
        query = partial(
            client.distinct,
            *metadata_keys,
            structure_families=structure_families,
            specs=specs,
            counts=counts,
        )
        return await loop.run_in_executor(None, query)


# Create a default catalog for basic usage
catalog = Catalog()


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.devices import slits


def test_slits_tweak():
    """Test the PVs for the tweak forward/reverse."""
    slits_obj = slits.BladeSlits("255idc:KB_slits", name="KB_slits")
    # Check the inherited setpoint/readback PVs
    assert slits_obj.v.center.setpoint.pvname == "255idc:KB_slitsVcenter"
    assert slits_obj.v.center.readback.pvname == "255idc:KB_slitsVt2.D"
    # Check the tweak PVs
    assert (
        slits_obj.v.center.tweak_value.pvname == "255idc:KB_slitsVcenter_tweakVal.VAL"
    )
    assert slits_obj.v.center.tweak_reverse.pvname == "255idc:KB_slitsVcenter_tweak.A"
    assert slits_obj.v.center.tweak_forward.pvname == "255idc:KB_slitsVcenter_tweak.B"


def test_aperture_PVs():
    aperture = slits.ApertureSlits(
        "255ida:slits:US:",
        pitch_motor="m1",
        yaw_motor="m2",
        horizontal_motor="m3",
        diagonal_motor="m4",
        name="whitebeam_slits",
    )
    assert not aperture.connected
    assert hasattr(aperture, "h")
    assert hasattr(aperture.h, "center")
    # Test pseudo motors controlled through the transform records
    assert aperture.h.center.user_readback.pvname == "255ida:slits:US:hCenter.RBV"
    assert aperture.v.center.user_readback.pvname == "255ida:slits:US:vCenter.RBV"
    assert aperture.h.size.user_readback.pvname == "255ida:slits:US:hSize.RBV"
    assert aperture.v.size.user_readback.pvname == "255ida:slits:US:vSize.RBV"
    # Test real motors
    assert aperture.pitch.user_readback.pvname == "255ida:slits:m1.RBV"
    assert aperture.yaw.user_readback.pvname == "255ida:slits:m2.RBV"
    assert aperture.horizontal.user_readback.pvname == "255ida:slits:m3.RBV"
    assert aperture.diagonal.user_readback.pvname == "255ida:slits:m4.RBV"
    # Check the derived signals are simple pass-throughs to the user readback/setpoint
    assert aperture.h.size.readback._derived_from.pvname == "255ida:slits:US:hSize.RBV"
    assert aperture.h.size.setpoint._derived_from.pvname == "255ida:slits:US:hSize.VAL"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest

from firefly.table import TableDisplay
from haven.devices import Table


@pytest.fixture()
async def table(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    tbl = Table(
        name="table",
        upstream_prefix="255idc:m1",
        downstream_prefix="255idc:m2",
        horizontal_prefix="255idc:m3",
        vertical_prefix="255idc:m2",
        pseudo_motor_prefix="255idc:table_ds:",
        transform_prefix="255idc:table_ds_trans:",
        labels={"tables"},
    )
    await tbl.connect(mock=True)
    return tbl


@pytest.fixture()
def empty_table(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    tbl = Table(
        name="table",
    )
    sim_registry.register(tbl)
    return tbl


@pytest.fixture()
def display(qtbot, table):
    disp = TableDisplay(macros={"DEVICE": table.name})
    qtbot.addWidget(disp)
    return disp


def test_unused_motor_widgets(qtbot, empty_table):
    """Do control widgets get disable for motors that aren't on the device?"""
    display = TableDisplay(macros={"DEVICE": empty_table.name})
    qtbot.addWidget(display)
    # Check that the bender control widgets were enabled
    assert not display.ui.pitch_embedded_display.isEnabled()
    assert not display.ui.vertical_embedded_display.isEnabled()
    assert not display.ui.horizontal_embedded_display.isEnabled()



import datetime as dt
import logging
from datetime import datetime
from unittest.mock import MagicMock
from zoneinfo import ZoneInfo

import pandas as pd
import pytest
import time_machine
from ophyd_async.core import set_mock_value
from tiled.adapters.mapping import MapAdapter
from tiled.adapters.xarray import DatasetAdapter
from tiled.client import Context, from_context
from tiled.server.app import build_app

from haven.devices import Motor
from haven.motor_position import (
    get_motor_position,
    get_motor_positions,
    list_current_motor_positions,
    list_motor_positions,
    recall_motor_position,
    save_motor_position,
)

log = logging.getLogger(__name__)

# Use a timezone we're not likely to be in for testing tz-aware behavior
fake_time = dt.datetime(2022, 8, 19, 19, 10, 51, tzinfo=ZoneInfo("Asia/Taipei"))


position_runs = {
    "a9b3e0fa-eba1-43e0-a38c-c7ac76278000": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(
                        pd.DataFrame(
                            {
                                "motor_A": [12.0],
                                "motor_B": [-113.25],
                            }
                        ).to_xarray()
                    ),
                },
                metadata={
                    "descriptors": [
                        {
                            "data_keys": {
                                "motor_A": {"object_name": "motor_A"},
                                "motor_B": {"object_name": "motor_B"},
                            },
                        }
                    ],
                },
            ),
        },
        metadata={
            "plan_name": "save_motor_position",
            "position_name": "Good position A",
            "time": 1725897133,
            "uid": "a9b3e0fa-eba1-43e0-a38c-c7ac76278000",
            "start": {
                "plan_name": "save_motor_position",
                "position_name": "Good position A",
                "time": 1725897133,
                "uid": "a9b3e0fa-eba1-43e0-a38c-c7ac76278000",
            },
        },
    ),
    # A second saved motor position
    "1b7f2ef5-6a3c-496e-9f6f-f1a4805c0065": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(
                        pd.DataFrame(
                            {
                                "motorC": [11250.0],
                            }
                        ).to_xarray()
                    ),
                },
                metadata={
                    "descriptors": [
                        {
                            "data_keys": {
                                "motorC": {"object_name": "motorC"},
                            },
                        }
                    ],
                },
            ),
        },
        metadata={
            "plan_name": "save_motor_position",
            "position_name": "A good position B",
            "time": 1725897193,
            "uid": "1b7f2ef5-6a3c-496e-9f6f-f1a4805c0065",
            "start": {
                "plan_name": "save_motor_position",
                "position_name": "A good position B",
                "time": 1725897193,
                "uid": "1b7f2ef5-6a3c-496e-9f6f-f1a4805c0065",
            },
        },
    ),
    # A saved motor position, but older
    "5dd9a185-d5c4-4c8b-a719-9d7beb9007dc": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(
                        pd.DataFrame(
                            {
                                "motorC": [11250.0],
                            }
                        ).to_xarray()
                    ),
                },
                metadata={
                    "descriptors": [
                        {
                            "data_keys": {
                                "motorC": {"object_name": "motorC"},
                            },
                        }
                    ],
                },
            ),
        },
        metadata={
            "plan_name": "save_motor_position",
            "position_name": "Good position C",
            "time": 1725897033,
            "uid": "5dd9a185-d5c4-4c8b-a719-9d7beb9007dc",
            "start": {
                "plan_name": "save_motor_position",
                "position_name": "Good position C",
                "time": 1725897033,
                "uid": "5dd9a185-d5c4-4c8b-a719-9d7beb9007dc",
            },
        },
    ),
    # A saved motor position, but older
    "42b8c45d-e98d-4f59-9ce8-8f14134c90bd": MapAdapter(
        {
            "primary": MapAdapter(
                {
                    "data": DatasetAdapter.from_dataset(
                        pd.DataFrame(
                            {
                                "motorC": [11250.0],
                            }
                        ).to_xarray()
                    ),
                },
                metadata={
                    "descriptors": [
                        {
                            "data_keys": {
                                "motorC": {"object_name": "motorC"},
                            },
                        }
                    ],
                },
            ),
        },
        metadata={
            "plan_name": "save_motor_position",
            "position_name": "Another good position",
            "time": 1725897233,
            "uid": "42b8c45d-e98d-4f59-9ce8-8f14134c90bd",
            "start": {
                "plan_name": "save_motor_position",
                "position_name": "Another good position",
                "time": 17258972333,
                "uid": "42b8c45d-e98d-4f59-9ce8-8f14134c90bd",
            },
        },
    ),
    # A scan that's not a saved motor position
    "9bcd07e9-3188-49d3-a1ce-e3b51ebe48b5": MapAdapter(
        {},
        metadata={
            "plan_name": "xafs_scan",
            "time": 1725897133,
        },
    ),
}


@pytest.fixture()
def client(mocker):
    tree = MapAdapter(position_runs)
    app = build_app(tree)
    with Context.from_app(app) as context:
        client = from_context(context)
        mocker.patch(
            "haven.motor_position.tiled_client", MagicMock(return_value=client)
        )
        yield client


@pytest.fixture
async def motors(sim_registry):
    # Create the motors
    motors = [
        Motor("", name="motor_A"),
        Motor("", name="motor_B"),
    ]
    for motor in motors:
        await motor.connect(mock=True)
        sim_registry.register(motor)
    return motors


def test_save_motor_position_by_device(motors):
    # Move to some other motor position so we can tell it saved the right one
    motorA, motorB = motors
    # Save the current motor position
    plan = save_motor_position(motorA, motorB, name="Sample center")
    # Check that the right read messages get emitted
    messages = list(plan)
    # Check that the motors got saved
    readA, readB = messages[6:8]
    assert readA.obj is motorA
    assert readB.obj is motorB


def test_save_motor_position_by_name(motors):
    # Check that no entry exists before saving it
    motorA, motorB = motors
    # Save the current motor position
    plan = save_motor_position(motorA.name, motorB.name, name="Sample center")
    # Check that the motors got saved
    # Check that the right read messages get emitted
    messages = list(plan)
    # Check that the motors got saved
    readA, readB = messages[6:8]
    assert readA.obj is motorA
    assert readB.obj is motorB


def test_get_motor_position(client):
    uid = "a9b3e0fa-eba1-43e0-a38c-c7ac76278000"
    result = get_motor_position(uid=uid)
    assert result.name == "Good position A"
    assert result.motors[0].name == "motor_A"
    assert result.motors[0].readback == 12.0


async def test_get_motor_positions(client):
    results = get_motor_positions(after=1725897100, before=1725897200)
    results = [pos async for pos in results]
    assert len(results) == 2
    # Check the motor position details
    motorA, motorB = results
    assert motorA.uid == "a9b3e0fa-eba1-43e0-a38c-c7ac76278000"


async def test_get_motor_positions_by_name(client):
    results = get_motor_positions(name=r"^.*good.+itio.+[AB]$", case_sensitive=False)
    results = [pos async for pos in results]
    print([r.name for r in results])
    assert len(results) == 2
    # Check the motor position details
    motorA, motorB = results
    assert motorA.uid == "a9b3e0fa-eba1-43e0-a38c-c7ac76278000"


def test_recall_motor_position(client, motors):
    # Re-set the previous value
    uid = "a9b3e0fa-eba1-43e0-a38c-c7ac76278000"
    plan = recall_motor_position(uid=uid)
    messages = list(plan)
    # Check the plan output
    msg0 = messages[0]
    assert msg0.obj.name == "motor_A"
    assert msg0.args[0] == 12.0
    msg1 = messages[1]
    assert msg1.obj.name == "motor_B"
    assert msg1.args[0] == -113.25


@time_machine.travel(fake_time, tick=True)
async def test_list_motor_positions(client, capsys):
    # Do the listing
    await list_motor_positions()
    # Check stdout for printed motor positions
    captured = capsys.readouterr()
    assert len(captured.out) > 0
    first_motor = captured.out.split("\n\n")[0]
    uid = "a9b3e0fa-eba1-43e0-a38c-c7ac76278000"
    timestamp = "2024-09-09 23:52:13"
    expected = "\n".join(
        [
            f"Good position A",
            f'┃ uid="{uid}", {timestamp}',
            f"┣━motor_A: 12.0, offset: None",
            f"┗━motor_B: -113.25, offset: None",
        ]
    )
    assert first_motor == expected


@time_machine.travel(fake_time, tick=True)
async def test_list_current_motor_positions(motors, capsys):
    # Get our simulated motors into the device registry
    motorA, motorB = motors
    with capsys.disabled():
        # Move to some other motor position so we can tell it saved the right one
        set_mock_value(motorA.user_readback, 11.0)
        set_mock_value(motorA.user_offset, 1.5)
        set_mock_value(motorB.user_readback, 23.0)
    # List the current motor position
    await list_current_motor_positions(motorA, motorB, name="Current motor positions")
    # Check stdout for printed motor positions
    captured = capsys.readouterr()
    assert len(captured.out) > 0
    timestamp = "2022-08-19 19:10:51"
    expected = "\n".join(
        [
            f"Current motor positions",
            f"┃ {timestamp}",
            f"┣━motor_A: 11.0, offset: 1.5",
            f"┗━motor_B: 23.0, offset: 0.0",
        ]
    )
    assert captured.out.strip("\n") == expected.strip("\n")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Ophyd support for the EPICS transform record."""

import asyncio

# from ophyd import Device
from ophyd_async.core import (
    ConfigSignal,
    Device,
    DeviceVector,
    HintedSignal,
    StandardReadable,
)
from ophyd_async.epics.signal import epics_signal_r, epics_signal_rw

from ..typing import StrEnum
from .synApps import EpicsRecordDeviceCommonAll, EpicsSynAppsRecordEnableMixin

CHANNEL_LETTERS_LIST = "A B C D E F G H I J K L M N O P".split()


#############################
# End common synApps support
#############################


class TransformRecordChannel(StandardReadable):
    """
    channel of a synApps transform record: A-P

    .. index:: Ophyd Device; synApps transformRecordChannel

    .. autosummary::

        ~reset
    """

    class PVValidity(StrEnum):
        EXT_PV_NC = "Ext PV NC"
        EXT_PV_OK = "Ext PV OK"
        LOCAL_PV = "Local PV"
        CONSTANT = "Constant"

    def __init__(self, prefix, letter, name=""):
        self._ch_letter = letter
        with self.add_children_as_readables():
            self.current_value = epics_signal_rw(float, f"{prefix}.{letter}")
        with self.add_children_as_readables(ConfigSignal):
            self.input_pv = epics_signal_rw(str, f"{prefix}.INP{letter}")
            self.comment = epics_signal_rw(str, f"{prefix}.CMT{letter}")
            self.expression = epics_signal_rw(
                str,
                f"{prefix}.CLC{letter}",
            )
        self.output_pv = epics_signal_rw(str, f"{prefix}.OUT{letter}")
        self.last_value = epics_signal_r(float, f"{prefix}.L{letter}")
        self.input_pv_valid = epics_signal_r(self.PVValidity, f"{prefix}.I{letter}V")
        self.expression_invalid = epics_signal_r(int, f"{prefix}.C{letter}V")
        self.output_pv_valid = epics_signal_r(
            self.PVValidity,
            f"{prefix}.O{letter}V",
        )

        super().__init__(name=name)

    async def reset(self):
        """set all fields to default values"""
        await asyncio.gather(
            self.comment.set(self._ch_letter.lower()),
            self.input_pv.set(""),
            self.expression.set(""),
            self.current_value.set(0),
            self.output_pv.set(""),
        )


class TransformRecord(EpicsRecordDeviceCommonAll):
    """
    EPICS transform record support in ophyd

    .. index:: Ophyd Device; synApps TransformRecord

    .. autosummary::

        ~reset

    :see: https://htmlpreview.github.io/?https://raw.githubusercontent.com/epics-modules/calc/R3-6-1/documentation/TransformRecord.html#Fields
    """

    class CalcOption(StrEnum):
        CONDITIONAL = "Conditional"
        ALWAYS = "Always"

    class InvalidLinkAction(StrEnum):
        IGNORE_ERROR = "Ignore error"
        DO_NOTHING = "Do Nothing"

    def __init__(self, prefix, name=""):
        with self.add_children_as_readables(ConfigSignal):
            self.units = epics_signal_rw(
                str,
                f"{prefix}.EGU",
            )
            self.precision = epics_signal_rw(int, f"{prefix}.PREC")
            self.version = epics_signal_r(
                float,
                f"{prefix}.VERS",
            )

            self.calc_option = epics_signal_rw(
                self.CalcOption,
                f"{prefix}.COPT",
            )
            self.invalid_link_action = epics_signal_r(
                self.InvalidLinkAction,
                f"{prefix}.IVLA",
            )
            self.input_bitmap = epics_signal_r(
                int, f"{prefix}.MAP", name="input_bitmap"
            )
        with self.add_children_as_readables():
            self.channels = DeviceVector(
                {
                    char: TransformRecordChannel(prefix=prefix, letter=char)
                    for char in CHANNEL_LETTERS_LIST
                }
            )

        super().__init__(prefix=prefix, name=name)
        # Remove dtype, it's broken for some reason
        del self.device_type

    async def reset(self):
        """set all fields to default values"""
        channels = self.channels.values()
        await asyncio.gather(
            self.scanning_rate.set(self.ScanInterval.PASSIVE),
            self.description.set(self.name),
            self.units.set(""),
            self.calc_option.set(0),
            self.precision.set(3),
            self.forward_link.set(""),
            *[ch.reset() for ch in channels],
        )
        # Restore the hinted channels
        self.add_readables(channels, HintedSignal)


class UserTransformN(EpicsSynAppsRecordEnableMixin, TransformRecord):
    """Single instance of the userTranN database."""


class UserTransformsDevice(Device):
    """
    EPICS synApps XXX IOC setup of userTransforms: ``$(P):userTran$(N)``

    .. index:: Ophyd Device; synApps UserTransformsDevice
    """

    def __init__(self, prefix, name=""):
        # Config attrs
        with self.add_children_as_readables(ConfigSignal):
            self.enable = epics_signal_rw(int, f"{prefix}userTranEnable", name="enable")
        # Read attrs
        with self.add_children_as_readables():
            self.transform1 = UserTransformN("userTran1")
            self.transform1 = UserTransformN("userTran2")
            self.transform1 = UserTransformN("userTran3")
            self.transform1 = UserTransformN("userTran4")
            self.transform1 = UserTransformN("userTran5")
            self.transform1 = UserTransformN("userTran6")
            self.transform1 = UserTransformN("userTran7")
            self.transform1 = UserTransformN("userTran8")
            self.transform1 = UserTransformN("userTran9")
            self.transform1 = UserTransformN("userTran10")

    async def reset(self):  # lgtm [py/similar-function]
        """set all fields to default values"""
        await asyncio.gather(
            self.transform1.reset(),
            self.transform2.reset(),
            self.transform3.reset(),
            self.transform4.reset(),
            self.transform5.reset(),
            self.transform6.reset(),
            self.transform7.reset(),
            self.transform8.reset(),
            self.transform9.reset(),
            self.transform10.reset(),
        )
        self.add_readables(
            [
                self.transform1,
                self.transform2,
                self.transform3,
                self.transform4,
                self.transform5,
                self.transform6,
                self.transform7,
                self.transform8,
                self.transform9,
                self.transform10,
            ]
        )


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: (c) 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



#!/usr/bin/env python3
from caproto.server import PVGroup, PvpropertyDouble, ioc_arg_parser, pvproperty, run

from haven.simulated_ioc import ResponsiveMotorFields  # , IOC as IOC_


class UndulatorGroup(PVGroup):
    """
    An IOC that looks like an undulator.

    E.g. "25IDds:Energy"

    """

    ScanEnergy = pvproperty(value=0, doc="ID Energy Scan Input", dtype=PvpropertyDouble)
    Energy = pvproperty(
        value=0,
        doc="",
        record=ResponsiveMotorFields,
        dtype=PvpropertyDouble,
    )
    Busy = pvproperty(value=0, doc="", name="Busy")
    Stop = pvproperty(value=0, doc="", name="Stop")
    Gap = pvproperty(value=0, doc="", name="Gap")
    TaperEnergy = pvproperty(value=0, doc="")
    TaperGap = pvproperty(value=0, doc="")
    Start = pvproperty(value=0, doc="")
    HarmonicValue = pvproperty(value=0, doc="")
    DeadbandGap = pvproperty(value=0, doc="")
    DeviceLimit = pvproperty(value=0, doc="")
    AccessSecurity = pvproperty(value=0, doc="")
    TotalPower = pvproperty(value=0, doc="")
    Message1 = pvproperty(value=0, doc="")
    Message2 = pvproperty(value=0, doc="")
    Message3 = pvproperty(value=0, doc="")
    ShClosedTime = pvproperty(value=0, doc="")
    Device = pvproperty(value=0, doc="")
    Location = pvproperty(value=0, doc="")
    Version = pvproperty(value=0, doc="")


if __name__ == "__main__":
    ioc_options, run_options = ioc_arg_parser(
        default_prefix="ID255:", desc="haven.tests.ioc_undulator test IOC"
    )
    ioc = UndulatorGroup(**ioc_options)
    run(ioc.pvdb, **run_options)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
